"""File paths module for SWMMAnywhere."""

from __future__ import annotations

import json
from collections import Counter
from pathlib import Path

from swmmanywhere.logging import logger
from swmmanywhere.utilities import yaml_dump, yaml_load


def next_directory(keyword: str, directory: Path) -> int:
    """Find the next directory number.

    Find the next directory number within a directory with a <keyword>_ in its
    name.

    Args:
        keyword (str): Keyword to search for in the directory name.
        directory (Path): Path to the directory to search within.

    Returns:
        int: Next directory number.
    """
    existing_dirs = [int(d.name.split("_")[-1]) for d in directory.glob(f"{keyword}_*")]
    return 1 if not existing_dirs else max(existing_dirs) + 1


def check_bboxes(bbox: tuple[float, float, float, float], data_dir: Path) -> int | bool:
    """Find the bounding box number.

    Check if the bounding box coordinates match any existing bounding box
    directories within data_dir.

    Args:
        bbox (tuple[float, float, float, float]): Bounding box coordinates in
            the format (minx, miny, maxx, maxy).
        data_dir (Path): Path to the data directory.

    Returns:
        int: Bounding box number if the coordinates match, else False.
    """
    # Find all bounding_box_info.json files
    info_fids = data_dir.glob("*/*bounding_box_info.json")

    # Iterate over info files
    for info_fid in info_fids:
        # Read bounding_box_info.json
        with info_fid.open("r") as info_file:
            bounding_info = json.load(info_file)
        # Check if the bounding box coordinates match
        if Counter(bounding_info.get("bbox")) == Counter(bbox):
            bbox_full_dir = info_fid.parent
            bbox_dir = bbox_full_dir.name
            bbox_number = int(bbox_dir.replace("bbox_", ""))
            return bbox_number

    return False


def get_next_bbox_number(
    bbox: tuple[float, float, float, float], data_dir: Path
) -> int:
    """Get the next bounding box number.

    If there are existing bounding box directories, check within them to see if
    any have the same bounding box, otherwise find the next number. If
    there are no existing bounding box directories, return 1.

    Args:
        bbox (tuple[float, float, float, float]): Bounding box coordinates in
            the format (minx, miny, maxx, maxy).
        data_dir (Path): Path to the data directory.

    Returns:
        int: Next bounding box number.
    """
    # Search for existing bounding box directories
    bbox_number = check_bboxes(bbox, data_dir)
    if not bbox_number:
        return next_directory("bbox", data_dir)
    return bbox_number


def get_overrides(klass: type, overrides: dict[str, Path]) -> dict[str, Path]:
    """Get overrides for a class."""
    out = {}
    for p in overrides.copy().keys():
        if not hasattr(klass, p):
            continue
        out[p] = overrides.pop(p)
    return out


class ProjectPaths:
    """Paths for the project folder (within base_dir)."""

    def __init__(
        self, base_dir: Path, project_name: str, extension: str = "parquet", **kwargs
    ):
        """Initialise the project paths.

        Args:
            base_dir (Path): The base directory.
            project_name (str): The name of the project.
            extension (str): The extension for the files.
            **kwargs: Additional file paths to override.
        """
        self.project_name = project_name
        self.extension = extension
        self.base_dir = base_dir
        self.overrides: dict[str, Path] = get_overrides(ProjectPaths, kwargs)

        self.project.mkdir(exist_ok=True)
        self.national.mkdir(exist_ok=True)

    @property
    def project(self):
        """The project folder (sits in the base_dir)."""
        return self.overrides.get("project", self.base_dir / self.project_name)

    @property
    def national(self):
        """The national folder (for national scale downloads)."""
        return self.overrides.get("national", self.project / "national")

    @property
    def national_building(self):
        """The national scale building file."""
        return self.overrides.get(
            "national_building", self.national / f"building.{self.extension}"
        )

    @property
    def whiteboxtools_binaries_zip(self):
        """The zipped whitebox tools binaries folder."""
        return self.overrides.get(
            "whiteboxtools_binaries_zip", self.project / "whitebox_tools.zip"
        )


class BBoxPaths:
    """Paths for the bounding box folder (within project folder)."""

    def __init__(
        self,
        project_paths: ProjectPaths,
        bbox_bounds: tuple[float, float, float, float],
        bbox_number: int | None = None,
        extension: str = "parquet",
        **kwargs,
    ):
        """Initialise the bounding box paths.

        Args:
            project_paths (ProjectPaths): The project paths.
            bbox_bounds (tuple[float, float, float, float]): Bounding box
                coordinates in the format (minx, miny, maxx, maxy).
            bbox_number (int, optional): The bounding box number. Defaults to
                None, in which case if the existing bounding box has already
                been created it will be found, otherwise the next number will
                be used.
            extension (str, optional): The extension for the files. Defaults to
                'parquet'.
            **kwargs: Additional file paths to override.
        """
        if not bbox_number:
            bbox_number = get_next_bbox_number(bbox_bounds, project_paths.project)

        self.base_dir = project_paths.project
        self.bbox_number = bbox_number
        self.extension = extension
        self.bbox_bounds = bbox_bounds
        self.overrides: dict[str, Path] = get_overrides(BBoxPaths, kwargs)

        self.bbox.mkdir(exist_ok=True)
        self.download.mkdir(exist_ok=True)

        bounding_box_info = {"bbox": bbox_bounds, "project": project_paths.project_name}

        bbox_info_file = self.bbox / "bounding_box_info.json"
        if not bbox_info_file.exists():
            with bbox_info_file.open("w") as info_file:
                json.dump(bounding_box_info, info_file, indent=2)

    @property
    def bbox(self):
        """The bounding box folder (specific to a bounding box)."""
        return self.overrides.get("bbox", self.base_dir / f"bbox_{self.bbox_number}")

    @property
    def download(self):
        """The download folder (for bbox specific downloaded data)."""
        return self.overrides.get("download", self.bbox / "download")

    @property
    def river(self):
        """The river graph for the bounding box."""
        return self.overrides.get("river", self.download / f"river.{self.extension}")

    @property
    def street(self):
        """The street graph for the bounding box."""
        return self.overrides.get("street", self.download / f"street.{self.extension}")

    @property
    def elevation(self):
        """The elevation file for the bounding box."""
        return self.overrides.get("elevation", self.download / "elevation.tif")

    @property
    def building(self):
        """The building file for the bounding box (clipped from national scale)."""
        return self.overrides.get(
            "building", self.download / f"building.geo{self.extension}"
        )

    @property
    def precipitation(self):
        """The precipitation data."""
        return self.overrides.get(
            "precipitation", self.download / f"precipitation.{self.extension}"
        )


class ModelPaths:
    """Paths for the model folder (within bbox folder)."""

    def __init__(
        self,
        bbox_paths: BBoxPaths,
        model_number: int | None = None,
        extension: str = "parquet",
        **kwargs,
    ):
        """Initialise the model paths.

        Args:
            bbox_paths (BBoxPaths): The bounding box paths.
            model_number (int, None): The model number. Defaults to None, in
                which case the next number in the bbox directory will be used.
            extension (str): The extension for the files.
            **kwargs: Additional file paths to override.
        """
        if model_number is None:
            model_number = next_directory("model", bbox_paths.bbox)

        self.base_dir = bbox_paths.bbox
        self.model_number = model_number
        self.extension = extension
        self.overrides: dict[str, Path] = get_overrides(ModelPaths, kwargs)

        self.model.mkdir(exist_ok=True)

    @property
    def model(self):
        """The model folder (one specific synthesised model)."""
        return self.overrides.get("model", self.base_dir / f"model_{self.model_number}")

    @property
    def inp(self):
        """The synthesised SWMM input file for the model."""
        return self.overrides.get("inp", self.model / f"model_{self.model_number}.inp")

    @property
    def subcatchments(self):
        """The subcatchments file for the model."""
        return self.overrides.get(
            "subcatchments", self.model / f"subcatchments.geo{self.extension}"
        )

    @property
    def graph(self):
        """The graph file for the model."""
        return self.overrides.get("graph", self.model / f"graph.{self.extension}")

    @property
    def nodes(self):
        """The nodes file for the model."""
        return self.overrides.get("nodes", self.model / f"nodes.geo{self.extension}")

    @property
    def edges(self):
        """The edges file for the model."""
        return self.overrides.get("edges", self.model / f"edges.geo{self.extension}")

    @property
    def streetcover(self):
        """The street cover file for the model."""
        return self.overrides.get(
            "streetcover", self.model / f"streetcover.geo{self.extension}"
        )


def filepaths_from_yaml(f: Path):
    """Get file paths from a yaml file."""
    address_dict = yaml_load(f.read_text())
    address_dict["base_dir"] = Path(address_dict["base_dir"])
    overrides = address_dict.pop("overrides", {})
    addresses = FilePaths(**address_dict, **overrides)
    return addresses


class FilePaths:
    """File paths class (manager for project, bbox and model)."""

    def __init__(
        self,
        base_dir: Path,
        project_name: str,
        bbox_bounds: tuple[float, float, float, float],
        bbox_number: int | None = None,
        model_number: int | None = None,
        extension: str = "parquet",
        **kwargs,
    ):
        """Initialise the file paths.

        Args:
            base_dir (Path): The base directory.
            project_name (str): The name of the project.
            bbox_bounds (tuple[float, float, float, float]): Bounding box
                coordinates in the format (minx, miny, maxx, maxy).
            bbox_number (int, optional): The bounding box number. Defaults to
                None, in which case if the existing bounding box has already
                been created it will be found, otherwise the next number will
                be used.
            model_number (int, optional): The model number. Defaults to None,
                in which case the next number in the bbox directory will be used.
            extension (str): The extension for the files.
            **kwargs: Additional file paths.
        """
        # Validate overrides and convert to paths
        for p, value in kwargs.items():
            value = Path(value)
            if not value.exists():
                logger.warning(f"Override path for {p}, {value} does not yet exist.")
            kwargs[p] = value

        # Create project paths and apply overrides
        self.project_paths = ProjectPaths(base_dir, project_name, extension, **kwargs)

        # Create bbox paths and apply overrides
        self.bbox_paths = BBoxPaths(
            self.project_paths, bbox_bounds, bbox_number, extension, **kwargs
        )

        # Create model paths and apply overrides
        self.model_paths = ModelPaths(
            self.bbox_paths, model_number, extension, **kwargs
        )

        self._overrides = kwargs

    def to_yaml(self, f: Path):
        """Convert a file to json."""
        address_dict = {}
        for attr in ["model_paths", "bbox_paths", "project_paths"]:
            address_dict.update(getattr(self, attr).__dict__)
            address_dict.update(**getattr(getattr(self, attr), "overrides"))
        address_dict.update(self._overrides)
        yaml_dump(address_dict, f.open("w"))

    def get_path(self, name: str) -> Path:
        """Get a path from _overrides."""
        path = self._overrides.get(name, None)
        if not path:
            raise FileExistsError(f"No file found for `{name}` attribute.")
        return path

    def set_bbox_number(self, number):
        """Set the bounding box number."""
        self.bbox_paths.bbox_number = number
        self.model_paths.base_dir = self.bbox_paths.bbox

    def set_model_number(self, number):
        """Set the model number."""
        self.model_paths.model_number = number



# %% [markdown]
# # Network comparison demo
# Note - this script can also be opened in interactive Python if you wanted to
# play around. On the GitHub it is in [docs/notebooks](https://github.com/ImperialCollegeLondon/SWMManywhere/blob/main/docs/notebooks/validation_demo.py)
# . To run this on your
# local machine, you will need to install the optional dependencies for `doc`:
#
# `pip install swmmanywhere[doc]`
#
# %% [markdown]
# ## Introduction
#
# This script demonstrates how to use SWMManywhere when you have a real network. It
# shows how to tell SWMManywhere where the necessary data is, and how, when this is
# provided, a suite of metrics are calculated to compare the real network with the
# synthesised network.
#
# Since this is a notebook, we will define [`config`](https://imperialcollegelondon.github.io/SWMManywhere/config_guide/)
# as a dictionary rather than a `yaml` file, but the same principles apply.
#
# ## Initial setup
#
# Here we will run one of the Bellinge sub-networks which is provided in the test data.
# We will keep everything in a temporary directory.
# %%
# Imports
from __future__ import annotations

import tempfile
from pathlib import Path

from swmmanywhere.defs import copy_test_data
from swmmanywhere.logging import set_verbose
from swmmanywhere.swmmanywhere import swmmanywhere
from swmmanywhere.utilities import plot_map

# Create temporary directory
temp_dir = tempfile.TemporaryDirectory(dir=".")
base_dir = Path(temp_dir.name)

# Make a folder for real data
real_dir = base_dir / "real"
real_dir.mkdir(exist_ok=True)

# Copy test data into the real data folder
copy_test_data(real_dir)

# %% [markdown]
# ## Create config file
#
# Below, we update our config file to use new coordinates, and provide the real data. As
# set out in the [schema](reference-defs.md#schema-for-configuration-file), a variety of
# data entries can be provided to describe the real network, all CRS of shapefiles must
# be in the UTM CRS for the project:
#
# - `graph` (essential) - a graph file in JSON format created from [`nx.node_link_data`](https://networkx.org/documentation/stable/reference/readwrite/generated/networkx.readwrite.json_graph.node_link_data.html)
# - `subcatchments` (essential) - a GeoJSON file with subcatchment outlines, with
# headings for `id`, and `impervious_area` (only needed if calculating the metric,
# `bias_flood_depth`).
# - `inp` - an inp file to run the model with, matching the appropriate graph
# and subcatchments.
# - `results` - a results file to compare the model with.
#
# At least one of `results` or `inp` must be provided. If `inp` is provided then the
# model will be run and the results compared to the real data, otherwise `results` will
# be loaded directly.
#
# *Note* that the need to separately provide a `graph` and `subcatchments` will be
# removed following fixing of [this](https://github.com/ImperialCollegeLondon/SWMManywhere/issues/84).
# %%
# Define config
config = {
    "base_dir": base_dir,
    "project": "bellinge_small",
    "bbox": [10.309, 55.333, 10.317, 55.339],
    "run_settings": {"duration": 3600},
    "real": {
        "graph": real_dir / "bellinge_small_graph.json",
        "inp": real_dir / "bellinge_small.inp",
        "subcatchments": real_dir / "bellinge_small_subcatchments.geojson",
    },
    "parameter_overrides": {
        "topology_derivation": {
            "allowable_networks": ["drive"],
            "omit_edges": ["bridge"],
        }
    },
}

# %% [markdown]
# ## Run SWMManywhere
#
# We make the `swmmanywhere` call as normal, but can observe that there is an additional
# model run (scroll towards the bottom) where the real model `inp` is being run and the
# metrics are shown to have been calculated in the log.
# %%
## Run SWMManywhere
set_verbose(True)
outputs = swmmanywhere(config)

# %% [markdown]
# ## Plot results
#
# We can plot the real network data and simulation (click on links for flow and nodes
# for flooding).
# %%
## View real data
plot_map(real_dir)

# %% [markdown]
# ... and we can plot the synthesised network
# %%
## View output
model_file = outputs[0]
plot_map(model_file.parent)

# %% [markdown]
# .. but of course we can see that the two networks do not perfectly line up. So we
# can't exactly plot our timeseries side-by-side. To quantify this properly we need to
# draw on SWMManywhere's ability to compare two networks that don't line up, which is
# done using the `metrics` output, automatically calculated when `real` data is
# provided.
# %%
# View metrics
metrics = outputs[1]
print(metrics)

# %% [markdown]
# For more information on using metrics see our [metrics guide](metrics_guide.md). To
# understand how to make use of such information, see our [paper](https://doi.org/10.1016/j.envsoft.2025.106358)
# for example.
#



# -*- coding: utf-8 -*-
"""Test the prepare_data module.

By default downloads themselves are mocked, but these can be enabled with the
following test command:

pytest -m downloads
"""

from __future__ import annotations

import tempfile
from pathlib import Path
from unittest import mock

import geopandas as gpd
import networkx as nx
import osmnx as ox
import pytest
import rasterio
import yaml
from geopy.geocoders import Nominatim

from swmmanywhere import prepare_data as downloaders


# Test get_country
@pytest.mark.downloads
def test_get_uk_download():
    """Check a UK point."""
    # Coordinates for London, UK
    x = -0.1276
    y = 51.5074

    result = downloaders.get_country(x, y)

    assert result[2] == "GB"
    assert result[3] == "GBR"


@pytest.mark.downloads
def test_get_us_download():
    """Check a US point."""
    x = -113.43318
    y = 33.81869

    result = downloaders.get_country(x, y)

    assert result[2] == "US"
    assert result[3] == "USA"


@pytest.mark.downloads
def test_building_downloader_download():
    """Check buildings are downloaded."""
    # Coordinates for small country (VAT)
    x = 7.41839
    y = 43.73205
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        temp_fid = Path(temp_dir) / "temp.parquet"
        # Download
        response = downloaders.download_buildings(temp_fid, x, y)
        # Check response
        assert response == 200

        # Check file exists
        assert temp_fid.exists(), "Buildings data file not found after download."

        # Load data
        gdf = gpd.read_parquet(temp_fid)

        # Make sure has some rows
        assert gdf.shape[0] > 0


@pytest.mark.downloads
def test_building_bbox_downloader_download():
    """Check buildings are downloaded."""
    # Coordinates for small country (VAT)
    bbox = (-0.17929, 51.49638, -0.17383, 51.49846)
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        temp_fid = Path(temp_dir) / "temp.parquet"
        # Download
        downloaders.download_buildings_bbox(temp_fid, bbox)

        # Check file exists
        assert temp_fid.exists(), "Buildings data file not found after download."

        # Load data
        gdf = gpd.read_parquet(temp_fid)

        # Make sure has some rows
        assert gdf.shape[0] > 0


@pytest.mark.downloads
def test_street_downloader_download():
    """Check streets are downloaded and a specific point in the graph."""
    bbox = (-0.17929, 51.49638, -0.17383, 51.49846)
    G = downloaders.download_street(bbox)

    # Not sure if they they are likely to change the osmid
    assert 26389449 in G.nodes


@pytest.mark.downloads
def test_river_downloader_download():
    """Check rivers are downloaded and a specific point in the graph."""
    bbox = (0.0402, 51.55759, 0.09825591114207548, 51.6205)
    G = downloaders.download_river(bbox)

    # Not sure if they they are likely to change the osmid
    assert 21473922 in G.nodes


@pytest.mark.downloads
def test_elevation_downloader_download():
    """Check elevation downloads, writes, contains data, and a known elevation."""
    bbox = (-0.17929, 51.49638, -0.17383, 51.49846)

    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        temp_fid = Path(temp_dir) / "temp.tif"

        # Download
        downloaders.download_elevation(temp_fid, bbox)

        # Check response
        assert temp_fid.exists(), "Elevation data file not found after download."

        # Load data
        with rasterio.open(temp_fid) as src:
            data = src.read(1)  # Reading the first band as an example

        # Make sure it has some values
        assert data.size > 0, "Elevation data should have some values."

        # Test some property of data (not sure if they may change this
        # data)
        assert data.max().max() > 25, "Elevation data should be higher."


@pytest.fixture
def setup_mocks():
    """Set up get_country mock for the tests."""
    # Mock for geolocator.reverse
    mock_location = mock.Mock()
    mock_location.raw = {"address": {"country_code": "gb"}}

    # Mock Nominatim
    nominatim_patch = mock.patch.object(
        Nominatim, "reverse", return_value=mock_location
    )
    # Mock yaml.safe_load
    yaml_patch = mock.patch.object(yaml, "safe_load", return_value={"GB": "GBR"})

    with nominatim_patch, yaml_patch:
        yield


def test_get_uk(setup_mocks):
    """Check a UK point."""
    # Coordinates for London, UK
    x = -0.1276
    y = 51.5074

    # Call get_country
    result = downloaders.get_country(x, y)

    assert result[2] == "GB"
    assert result[3] == "GBR"


def test_building_downloader(setup_mocks):
    """Check buildings are downloaded."""
    # Coordinates
    x = -0.1276
    y = 51.5074

    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        temp_fid = Path(temp_dir) / "temp.parquet"
        mock_response = mock.Mock()
        mock_response.status_code = 200
        mock_response.content = b'{"features": []}'
        with mock.patch("requests.get", return_value=mock_response) as mock_get:
            # Call your function that uses requests.get
            response = downloaders.download_buildings(temp_fid, x, y)

            # Assert that requests.get was called with the right arguments
            mock_get.assert_called_once_with(
                "https://data.source.coop/vida/google-microsoft-open-buildings/geoparquet/by_country/country_iso=GBR/GBR.parquet"
            )

        # Check response
        assert response == 200


def test_street_downloader():
    """Check streets are downloaded and a specific point in the graph."""
    bbox = (-0.17929, 51.49638, -0.17383, 51.49846)

    mock_graph = nx.MultiDiGraph()
    # Mock ox.graph_from_bbox
    with mock.patch.object(ox, "graph_from_bbox", return_value=mock_graph):
        # Call download_street
        G = downloaders.download_street(bbox)
        assert G == mock_graph


def test_river_downloader():
    """Check rivers are downloaded and a specific point in the graph."""
    bbox = (0.0402, 51.55759, 0.09825591114207548, 51.6205)

    mock_graph = nx.MultiDiGraph()
    mock_graph.add_node(1)
    # Mock ox.graph_from_bbox
    with mock.patch.object(
        ox, "graph_from_bbox", return_value=mock_graph
    ) as mock_from_bbox:
        # Call download_street
        G = downloaders.download_river(bbox)
        assert G == mock_graph

        mock_from_bbox.side_effect = ValueError(
            "Found no graph nodes within the requested polygon"
        )
        G = downloaders.download_river(bbox)
        assert G.size() == 0


def test_download_elevation():
    """Check elevation downloads, writes, contains data, and a known elevation."""
    bbox = (-0.17929, 51.49638, -0.17383, 51.49846)

    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        temp_fid = Path(temp_dir) / "temp.tif"

        # Mock the external dependencies
        module_base = "swmmanywhere.prepare_data."
        with (
            mock.patch(f"{module_base}pystac_client.Client.open") as mock_open,
            mock.patch(f"{module_base}planetary_computer.sign_inplace"),
            mock.patch(f"{module_base}planetary_computer.sign") as mock_sign,
            mock.patch(f"{module_base}rioxarray.open_rasterio") as mock_open_rasterio,
            mock.patch(f"{module_base}rxr_merge.merge_arrays") as mock_merge_arrays,
        ):
            # Mock the behavior of the catalog search and items
            mock_catalog = mock.MagicMock()
            mock_open.return_value = mock_catalog
            mock_search = mock.MagicMock()
            mock_catalog.search.return_value = mock_search
            mock_items = [mock.MagicMock(), mock.MagicMock()]
            for item in mock_items:
                item.assets = {"elevation": mock.MagicMock()}
            mock_search.items.return_value = mock_items

            # Mock the signed URLs
            mock_sign.side_effect = lambda x: mock.MagicMock(href=f"signed_{x}")

            # Mock the raster data
            mock_raster = mock.MagicMock()
            mock_open_rasterio.return_value = mock_raster

            # Mock the merged array
            mock_merged_array = mock.MagicMock()
            mock_merge_arrays.return_value = mock_merged_array

            # Mock the `rio` attribute on the merged array
            mock_merged_array.rio = mock.MagicMock()
            mock_merged_array.rio.clip_box.return_value = mock_merged_array
            mock_merged_array.rio.to_raster.return_value = None

            # Call the function
            downloaders.download_elevation(temp_fid, bbox)

            # Assertions
            mock_open.assert_called_once_with(
                "https://planetarycomputer.microsoft.com/api/stac/v1",
                modifier=mock.ANY,
            )
            mock_catalog.search.assert_called_once_with(
                collections=["nasadem"],
                bbox=bbox,
            )
            assert len(mock_items) == 2
            assert mock_sign.call_count == len(mock_items)
            mock_open_rasterio.assert_called()
            mock_merge_arrays.assert_called_once()
            mock_merged_array.rio.clip_box.assert_called_once_with(*bbox)
            mock_merged_array.rio.to_raster.assert_called_once_with(temp_fid)



"""Module for graphfcns that design the pipe inverts and diameters."""

from __future__ import annotations

from itertools import product
from typing import Hashable

import networkx as nx
import numpy as np
import pandas as pd
from tqdm import tqdm

from swmmanywhere import parameters
from swmmanywhere.graph_utilities import BaseGraphFunction, register_graphfcn
from swmmanywhere.logging import logger, verbose


@register_graphfcn
class pipe_by_pipe(
    BaseGraphFunction,
    required_edge_attributes=["length"],
    required_node_attributes=["contributing_area", "surface_elevation"],
    adds_edge_attributes=["diameter"],
    adds_node_attributes=["chamber_floor_elevation"],
):
    """pipe_by_pipe class."""

    # If doing required_graph_attributes - it would be something like 'dendritic'

    def __call__(
        self, G: nx.Graph, hydraulic_design: parameters.HydraulicDesign, **kwargs
    ) -> nx.Graph:
        """Pipe by pipe hydraulic design.

        Starting from the most upstream node, design a pipe to the downstream node
        specifying a diameter and downstream invert level. A range of diameters and
        invert levels are tested (ranging between conditions defined in
        hydraulic_design). From the tested diameters/inverts, a selection is made based
        on each pipe's satisfying feasibility constraints on: surcharge velocity,
        filling ratio, (and shear stress - not currently implemented). Prioritising
        feasibility in this order it identifies pipes with the preferred feasibility
        level. If multiple pipes are feasible, it picks the lowest cost pipe. Once
        the feasible pipe is identified, the diameter and downstream invert are set
        and then the next downstream pipe can be designed.

        This approach is based on the pipe-by-pipe design proposed in:
            https://doi.org/10.1016/j.watres.2021.117903

        Args:
            G (nx.Graph): A graph
            hydraulic_design (parameters.HydraulicDesign): A HydraulicDesign parameter
                object
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            G (nx.Graph): A graph
        """
        G = G.copy()
        surface_elevations = nx.get_node_attributes(G, "surface_elevation")
        topological_order = list(nx.topological_sort(G))
        chamber_floor = {}

        edge_designs: dict[str, dict[tuple[Hashable, Hashable, int], float]] = {
            parameter: {} for parameter in hydraulic_design.edge_design_parameters
        }

        # Iterate over nodes in topological order
        for node in tqdm(topological_order, disable=not verbose()):
            # Check if there's any nodes upstream, if not set the depth to min_depth
            if len(nx.ancestors(G, node)) == 0:
                chamber_floor[node] = (
                    surface_elevations[node] - hydraulic_design.min_depth
                )

            self.process_successors(
                G,
                node,
                surface_elevations,
                chamber_floor,
                edge_designs,
                hydraulic_design,
            )

        for parameter in hydraulic_design.edge_design_parameters:
            nx.function.set_edge_attributes(G, edge_designs[parameter], parameter)

        nx.function.set_node_attributes(G, chamber_floor, "chamber_floor_elevation")
        return G

    @staticmethod
    def get_designs(hydraulic_design: parameters.HydraulicDesign) -> product:
        """Get the designs for the pipe.

        This function generates a grid of designs for the pipe based on the
        diameters and depths specified in the hydraulic design parameters. It
        returns an iterable product of the designs.

        Args:
            hydraulic_design (parameters.HydraulicDesign): A HydraulicDesign parameter
                object

        Returns:
            product: An iterable product object containing the designs
        """
        return product(
            hydraulic_design.diameters,
            np.linspace(
                hydraulic_design.min_depth,
                hydraulic_design.max_depth,
                hydraulic_design.depth_nbins,
            ),
        )

    @staticmethod
    def calculate_cost(V: float, diam: float) -> float:
        """Calculate the cost of the pipe.

        This function calculates the cost of the pipe based on the excavation volume and
        diameter.

        Cost equation from: https://doi.org/10.2166/hydro.2016.105

        Args:
            V (float): The excavation volume of the pipe
            diam (float): The diameter of the pipe
        Returns:
            float: The cost of the pipe in USD
        """
        return 1.32 / 2000 * (9579.31 * diam**0.5737 + 1163.77 * V**1.31)

    def evaluate_design(
        self,
        ds_elevation: float,
        chamber_floor: float,
        edge_length: float,
        hydraulic_design: parameters.HydraulicDesign,
        Q: float,
        diam: float,
        depth: float,
    ) -> dict[Hashable, float]:
        """Evaluate the design of a pipe.

        This function evaluates the design of a pipe by calculating the cost,
        velocity, filling ratio, and feasibility parameters. It returns a dictionary
        containing the design parameters and their values.

        Args:
            ds_elevation (float): The downstream elevation
            chamber_floor (float): The elevation of the chamber floor
            edge_length (float): The length of the edge
            hydraulic_design (parameters.HydraulicDesign): A HydraulicDesign parameter
                object
            Q (float): The flow rate
            diam (float): The diameter of the pipe
            depth (float): The depth of the pipe

        Returns:
            dict: A dictionary containing the designed pipe's parameters and values
        """
        A = np.pi * diam**2 / 4
        n = 0.012  # mannings n
        R = A / (np.pi * diam)  # hydraulic radius
        # TODO... presumably need to check depth > (diam + min_depth)

        elev_diff = chamber_floor - (ds_elevation - depth)
        slope = elev_diff / edge_length
        # Always pick a pipe that is feasible without surcharging
        # if available
        surcharge_feasibility = 0.0
        # Use surcharged elevation
        while slope <= 0:
            surcharge_feasibility += 0.05
            slope = (
                chamber_floor + surcharge_feasibility - (ds_elevation - depth)
            ) / edge_length
            # TODO could make the feasibility penalisation increase
            # when you get above surface_elevation[node]... but
            # then you'd need a feasibility tracker and an offset
            # tracker
        v = (slope**0.5) * (R ** (2 / 3)) / n
        filling_ratio = Q / (v * A)
        # buffers from: https://www.polypipe.com/sites/default/files/Specification_Clauses_Underground_Drainage.pdf
        average_depth = (depth + chamber_floor) / 2
        V = edge_length * (diam + 0.3) * (average_depth + 0.1)
        cost = self.calculate_cost(V, diam)
        v_feasibility = max(hydraulic_design.min_v - v, 0) + max(
            v - hydraulic_design.max_v, 0
        )
        fr_feasibility = max(filling_ratio - hydraulic_design.max_fr, 0)
        """
        TODO shear stress... got confused here
        density = 1000
        dyn_visc = 0.001
        hydraulic_diameter = 4 * (A * filling_ratio**2) / \
            (np.pi * diam * filling_ratio)
        Re = density * v * 2 * (diam / 4) * (filling_ratio ** 2) / dyn_visc
        fd = 64 / Re
        shear_stress = fd * density * v**2 / fd
        shear_feasibility = max(min_shear - shear_stress, 0)
        """
        slope = (chamber_floor - (ds_elevation - depth)) / edge_length
        return {
            "diameter": diam,
            "depth": depth,
            "slope": slope,
            "v": v,
            "fr": filling_ratio,
            # 'tau' : shear_stress,
            "cost_usd": cost,
            "v_feasibility": v_feasibility,
            "fr_feasibility": fr_feasibility,
            "surcharge_feasibility": surcharge_feasibility,
            # 'shear_feasibility' : shear_feasibility
        }

    def select_design(self, pipes_df: pd.DataFrame) -> dict[Hashable, float]:
        """Select the ideal design from the dataframe.

        This function selects the ideal design from the dataframe by sorting the
        dataframe based on the feasibility parameters and cost. It returns the
        diameter, depth, and cost of the ideal design.

        Args:
            pipes_df (pd.DataFrame): A dataframe containing the designs and their
                parameters

        Returns:
            dict: A dictionary containing the ideal design
        """
        if pipes_df.shape[0] <= 0:
            raise ValueError("No non nan pipes designed. Shouldn't happen.")

        ideal_pipe = pipes_df.sort_values(
            by=[
                "surcharge_feasibility",
                "v_feasibility",
                "fr_feasibility",
                # 'shear_feasibility',
                "depth",
                "cost_usd",
            ],
            ascending=True,
        ).iloc[0]

        return ideal_pipe.to_dict()

    def design_pipe(
        self,
        ds_elevation: float,
        chamber_floor: float,
        edge_length: float,
        hydraulic_design: parameters.HydraulicDesign,
        Q: float,
    ) -> dict[Hashable, float]:
        """Design a pipe.

        This function designs a pipe by iterating over a range of diameters and
        depths. It returns the diameter and depth of the pipe that minimises the
        cost function, while also maintaining or minimising feasibility parameters
        associated with: surcharging, velocity and filling ratio.

        Args:
            ds_elevation (float): The downstream elevationq
            chamber_floor (float): The elevation of the chamber floor
            edge_length (float): The length of the edge
            hydraulic_design (parameters.HydraulicDesign): A HydraulicDesign parameter
                object
            Q (float): The flow rate

        Returns:
            dict: A dictionary containing the selected design
        """
        # Generate designs
        designs = self.get_designs(hydraulic_design)

        # Evaluate designs
        pipes = [
            self.evaluate_design(
                ds_elevation,
                chamber_floor,
                edge_length,
                hydraulic_design,
                Q,
                diam,
                depth,
            )
            for diam, depth in designs
        ]

        # Return selected design
        return self.select_design(pd.DataFrame(pipes).dropna())

    @staticmethod
    def calculate_flow(G: nx.Graph, node: Hashable, design_precipitation) -> float:
        """Calculate the flow to a node.

        This function calculates the flow to a node by summing the flow from its
        predecessors. It returns the total flow to the node.

        Args:
            G (nx.Graph): A graph
            node (Hashable): A node
            design_precipitation (float): The design precipitation

        Returns:
            float: The total flow to the node
        """
        # Find contributing area with ancestors
        # TODO - could do timearea here if i hated myself enough
        anc = nx.ancestors(G, node).union([node])
        tot = sum([G.nodes[anc_node]["contributing_area"] for anc_node in anc])
        M3_PER_HR_TO_M3_PER_S = 1 / 60 / 60

        return tot * design_precipitation * M3_PER_HR_TO_M3_PER_S

    def process_successors(
        self,
        G: nx.Graph,
        node: Hashable,
        surface_elevations: dict[Hashable, float],
        chamber_floor: dict[Hashable, float],
        edge_designs: dict[str, dict[tuple[Hashable, Hashable, int], float]],
        hydraulic_design: parameters.HydraulicDesign,
    ) -> None:
        """Process the successors of a node.

        This function processes the successors of a node. It designs a pipe to the
        downstream node and sets the diameter and downstream invert level of the
        pipe. It also sets the downstream invert level of the downstream node. It
        returns None but modifies the hydraulic_design.edge_design_parameters entries
        inside edge_designs and node chamber_floor dictionaries.

        Args:
            G (nx.Graph): A graph
            node (Hashable): A node
            surface_elevations (dict): A dictionary of surface elevations keyed by
                node
            chamber_floor (dict): A dictionary of chamber floor elevations keyed by
                node
            edge_designs (dict): A dictionary of pipe designs keyed by parameter and
                then edge
            hydraulic_design (parameters.HydraulicDesign): A HydraulicDesign parameter
                object
        """
        Q = self.calculate_flow(G, node, hydraulic_design.precipitation)

        for ix, ds_node in enumerate(G.successors(node)):
            edge = G.get_edge_data(node, ds_node, 0)

            # Design the pipe to find the diameter and invert depth
            pipe = self.design_pipe(
                surface_elevations[ds_node],
                chamber_floor[node],
                edge["length"],
                hydraulic_design,
                Q,
            )

            for parameter in hydraulic_design.edge_design_parameters:
                edge_designs[parameter][(node, ds_node, 0)] = pipe[parameter]

            chamber_floor[ds_node] = min(
                surface_elevations[ds_node] - pipe["depth"],
                chamber_floor.get(ds_node, np.inf),
            )
            if ix > 0:
                logger.warning(
                    """a node has multiple successors, 
                    not sure how that can happen if using shortest path
                    to derive topology"""
                )



"""NetComp v0.2.2
==============

    NetComp is a Python package for comparing networks using pairwise distances,
    and for performing anomaly detection on a time series of networks. It is
    built on top of the NetworkX package. For details on usage and installation,
    see README.md.

"""

from __future__ import annotations

__author__ = "Peter Wills <peter.e.wills@gmail.com>"
__version__ = "0.2.2"
__license__ = "MIT"

import sys

if sys.version_info[0] < 3:
    m = "Python 3.x required (%d.%d detected)."
    raise ImportError(m % sys.version_info[:2])
del sys

from .linalg import *  # noqa
from .distance import *  # noqa



# %% [markdown]
# # Extended Demo
# Note - this script can also be opened in interactive Python if you wanted to
# play around. On the GitHub it is in [docs/notebooks](https://github.com/ImperialCollegeLondon/SWMManywhere/blob/main/docs/notebooks/extended_demo.py)
# . To run this on your
# local machine, you will need to install the optional dependencies for `doc`:
#
# `pip install swmmanywhere[doc]`
#
# %% [markdown]
# ## Introduction
# This script demonstrates a simple use case of `swmmanywhere`, building on the
# [quickstart](https://imperialcollegelondon.github.io/SWMManywhere/quickstart/)
# example, but including plotting and alterations.
#
# Since this is a notebook, we will define [`config`](https://imperialcollegelondon.github.io/SWMManywhere/config_guide/)
# as a dictionary rather than a `yaml` file, but the same principles apply.
#
# ## Initial run
#
# Here we will run the [quickstart](https://imperialcollegelondon.github.io/SWMManywhere/config_guide/)
# configuration, keeping everything in a temporary directory.
# %%
# Imports
from __future__ import annotations

import tempfile
from pathlib import Path

import folium
import geopandas as gpd
import pandas as pd

from swmmanywhere.logging import set_verbose
from swmmanywhere.swmmanywhere import swmmanywhere
from swmmanywhere.utilities import plot_map

# Create temporary directory
temp_dir = tempfile.TemporaryDirectory(dir=".")
base_dir = Path(temp_dir.name)

# Define minimum viable config (with shorter duration so better inspect results)
config = {
    "base_dir": base_dir,
    "project": "my_first_swmm",
    "bbox": [1.52740, 42.50524, 1.54273, 42.51259],
    "run_settings": {"duration": 3600},
}

# Run SWMManywhere
outputs = swmmanywhere(config)

# Verify the output
model_file = outputs[0]
if not model_file.exists():
    raise FileNotFoundError(f"Model file not created: {model_file}")


# %% [markdown]
# ## Plotting output
#
# If you do not have a real UDM, the majority of your interpretation will be
# around the synthesised `nodes` and `edges`. These are
# created in the same directory as the `model_file`. Let's have a look at them.
# Note that the `outfall` that each node drains to is specified in the `outfall`
# attribute, we will plot these in red and other nodes in black with the built-
# in `swmmanywhere.utilities.plot_map` function.
# %%
# Create a folium map and add the nodes and edges
plot_map(model_file.parent)

# %% [markdown]
# OK, it's done something! Though perhaps we're not super satisfied with the output.
#
# ## Customising outputs
#
# Some things stick out on first glance,
#
# - Probably we do not need pipes in the hills to the South, these seem to be along
# pedestrian routes, which can be adjusted with the `allowable_networks` parameter.
# - We will also remove any types under the `omit_edges` entry, here you can specify
# to not allow pipes to cross bridges, tunnels, motorways, etc., however, this is
# such a small area we probably don't want to restrict things so much.
# - We have far too few outfalls, it seems implausible that so many riverside streets
# would not have outfalls. Furthermore, there are points that are quite far from the
# river that have been assigned as outfalls. We can reduce the `river_buffer_distance`
# to make nodes nearer the river more likely to be outfalls, but also reduce the
# `outfall_length` distance parameter to enable `swmmanywhere` to more freely select
# outfalls that are adjacent to the river.
#
# Let's just demonstrate that using the
# [`parameter_overrides` functionality](https://imperialcollegelondon.github.io/SWMManywhere/config_guide/#changing-parameters).
#
# %%
config["parameter_overrides"] = {
    "topology_derivation": {
        "allowable_networks": ["drive"],
        "omit_edges": ["bridge"],
    },
    "outfall_derivation": {
        "outfall_length": 5,
        "river_buffer_distance": 30,
    },
}
outputs = swmmanywhere(config)
plot_map(outputs[0].parent)

# %% [markdown]
# OK that clearly helped, although we have appear to have stranded pipes (e.g., along
# *Carrer dels Canals* in North West), presumably due to some mistake in the
# OSM specifying that it
# is connected via a pedestrian route. We won't remedy this in the tutorial, but you can
# manually provide your
# [`starting_graph`](https://imperialcollegelondon.github.io/SWMManywhere/config_guide/#change-starting_graph)
# via the configuration file to address such mistakes.
#
# More importantly we can see some distinctive unconnected network in the South West.
# What is going on there? To explain this we will have to turn on verbosity to print the
# intermediate files used in model derivation.
#
# To do this with a command line call we simply add the flag `--verbose=True`.
# Though in code we will have to use `set_verbose` from the `logging` module.

# %%
# Make verbose
set_verbose(True)  # Set verbosity

# Run again
outputs = swmmanywhere(config)
model_dir = outputs[0].parent
m = plot_map(model_dir)

# %% [markdown]
# That's a lot of information! However, the reason we are currently interested
# in this is because the files associated with
# each workflow step are saved when `verbose=True`.
#
# We will load a file called `subbasins` and add it to the map.

# %%
subbasins = gpd.read_file(model_dir / "subbasins.geoparquet")
folium.GeoJson(subbasins, fill_opacity=0, color="blue", weight=2).add_to(m)
m

# %% [markdown]
# Although this can be customised, the default behaviour of `swmmanywhere` is to not
# allow edges to cross hydrological subbasins. It is now super clear why these
# unconnected  networks have appeared, and are ultimately due to the underlying DEM.
# If you did desperately care about these streets, then you should probably
# widen your bounding box.
#
# ## Plotting results
#
# Because we have run the model with `verbose=True` we will also see that a new
# `results` file has appeared, which contains all of the simulation results from SWMM.

# %%
df = pd.read_parquet(model_dir / "results.parquet")
df.head()

# %% [markdown]
# `results` contains all simulation results in long format, with `flooding` at
# nodes and `flow` at edges. We will plot a random `flow`.

# %%
flows = df.loc[df.variable == "flow"]
flows.loc[flows.id == flows.iloc[0].id].set_index("date").value.plot(
    ylabel="flow (l/s)"
)


# %% [markdown]
# If `results` are present in the `model_dir`, `plot_map` will make clickable
# elements to view plots,
# now you can inspect your results in a much more elegant way than the SWMM GUI.
# Just click a node or link to view the flooding or flow timeseries!


# %%
plot_map(model_dir)

# %% [markdown]
# If we explore around, clicking on edges, we can see that flows are often
# looking sensible, though we can definitely some areas that have been hampered
# by our starting street graph (e.g., in the Western portion of *Carrer del Sant Andreu*
# in North West we can see negative flows meaning the direction is different from
# what the topology derivation assumed flow would be going in!).
# The first suggestion here would be to examine the starting graph,
# however, if you want to
# make more sophisticated customisations then your probably want to learn about
# [graph functions](https://imperialcollegelondon.github.io/SWMManywhere/graphfcns_guide/).



"""***************
Exact Distances
***************

Calculation of exact distances between graphs. Generally slow (quadratic in
graph size).
"""

from __future__ import annotations

import networkx as nx
import numpy as np
from numpy import linalg as la

from ..exception import InputError
from ..linalg import (  # type: ignore
    fast_bp,
    normalized_laplacian_eig,
    renormalized_res_mat,
    resistance_matrix,
)
from ..linalg.eigenstuff import _eigs, _flat
from ..linalg.matrices import _pad, laplacian_matrix
from .features import aggregate_features, get_features

######################
## Helper Functions ##
######################


def _canberra_dist(v1, v2):
    """The canberra distance between two vectors. We need to carefully handle
    the case in which both v1 and v2 are zero in a certain dimension.
    """
    eps = 10 ** (-15)
    v1, v2 = [_flat(v) for v in [v1, v2]]
    d_can = 0
    for u, w in zip(v1, v2):
        if np.abs(u) < eps and np.abs(w) < eps:
            d_update = 1
        else:
            d_update = np.abs(u - w) / (np.abs(u) + np.abs(w))
        d_can += d_update
    return d_can


#############################
## Distance Between Graphs ##
#############################


def edit_distance(A1, A2):
    """The edit distance between graphs, defined as the number of changes one
    needs to make to put the edge lists in correspondence.

    Parameters
    ----------
    A1, A2 : NumPy matrices
        Adjacency matrices of graphs to be compared

    Returns:
    -------
    dist : float
        The edit distance between the two graphs
    """
    dist = np.abs((A1 - A2)).sum() / 2
    return dist


def vertex_edge_overlap(A1, A2):
    """Vertex-edge overlap. Basically a souped-up edit distance, but in similarity
    form. The VEO similarity is defined as

        VEO(G1,G2) = (|V1&V2| + |E1&E2|) / (|V1|+|V2|+|E1|+|E2|)

    where |S| is the size of a set S and U&T is the union of U and T.

    Parameters
    ----------
    A1, A2 : NumPy matrices
        Adjacency matrices of graphs to be compared

    Returns:
    -------
    sim : float
        The similarity between the two graphs


    References:
    ----------

    """
    try:
        [G1, G2] = [nx.from_scipy_sparse_array(A) for A in [A1, A2]]
    except AttributeError:
        [G1, G2] = [nx.from_numpy_array(A) for A in [A1, A2]]
    V1, V2 = [set(G.nodes()) for G in [G1, G2]]
    E1, E2 = [set(G.edges()) for G in [G1, G2]]
    V_overlap = len(V1 | V2)  # set union
    E_overlap = len(E1 | E2)
    sim = (V_overlap + E_overlap) / (len(V1) + len(V2) + len(E1) + len(E2))
    return sim


def vertex_edge_distance(A1, A2):
    """Vertex-edge overlap transformed into a distance via

        D = (1-VEO)/VEO

    which is the inversion of the common distance-to-similarity function

        sim = 1/(1+D).

    Parameters
    ----------
    A1, A2 : NumPy matrices
        Adjacency matrices of graphs to be compared

    Returns:
    -------
    dist : float
        The distance between the two graphs
    """
    sim = vertex_edge_overlap(A1, A2)
    dist = (1 - sim) / sim
    return dist


def lambda_dist(A1, A2, k=None, p=2, kind="laplacian"):
    """The lambda distance between graphs, which is defined as

        d(G1,G2) = norm(L_1 - L_2)

    where L_1 is a vector of the top k eigenvalues of the appropriate matrix
    associated with G1, and L2 is defined similarly.

    Parameters
    ----------
    A1, A2 : NumPy matrices
        Adjacency matrices of graphs to be compared

    k : Integer
        The number of eigenvalues to be compared

    p : non-zero Float
        The p-norm is used to compare the resulting vector of eigenvalues.

    kind : String , in {'laplacian','laplacian_norm','adjacency'}
        The matrix for which eigenvalues will be calculated.

    Returns:
    -------
    dist : float
        The distance between the two graphs

    Notes:
    -----
    The norm can be any p-norm; by default we use p=2. If p<0 is used, the
    result is not a mathematical norm, but may still be interesting and/or
    useful.

    If k is provided, then we use the k SMALLEST eigenvalues for the Laplacian
    distances, and we use the k LARGEST eigenvalues for the adjacency
    distance. This is because the corresponding order flips, as L = D-A.

    References:
    ----------

    See Also:
    --------
    netcomp.linalg._eigs
    normalized_laplacian_eigs

    """
    # ensure valid k
    n1, n2 = [A.shape[0] for A in [A1, A2]]
    N = min(n1, n2)  # minimum size between the two graphs
    if k is None or k > N:
        k = N
    if kind == "laplacian":
        # form matrices
        L1, L2 = [laplacian_matrix(A) for A in [A1, A2]]
        # get eigenvalues, ignore eigenvectors
        evals1, evals2 = [_eigs(L)[0] for L in [L1, L2]]
    elif kind == "laplacian_norm":
        # use our function to graph evals of normalized laplacian
        evals1, evals2 = [normalized_laplacian_eig(A)[0] for A in [A1, A2]]
    elif kind == "adjacency":
        evals1, evals2 = [_eigs(A)[0] for A in [A1, A2]]
        # reverse, so that we are sorted from large to small, since we care
        # about the k LARGEST eigenvalues for the adjacency distance
        evals1, evals2 = [evals[::-1] for evals in [evals1, evals2]]
    else:
        raise InputError(
            "Invalid type, choose from 'laplacian', 'laplacian_norm', and 'adjacency'."
        )
    dist = la.norm(evals1[:k] - evals2[:k], ord=p)
    return dist


def netsimile(A1, A2):
    """NetSimile distance between two graphs.

    Parameters
    ----------
    A1, A2 : SciPy sparse array
        Adjacency matrices of the graphs in question.

    Returns:
    -------
    d_can : Float
        The distance between the two graphs.

    Notes:
    -----
    NetSimile works on graphs without node correspondence. Graphs to not need to
    be the same size.

    See Also:
    --------

    References:
    ----------
    """
    feat_A1, feat_A2 = [get_features(A) for A in [A1, A2]]
    agg_A1, agg_A2 = [aggregate_features(feat) for feat in [feat_A1, feat_A2]]
    # calculate Canberra distance between two aggregate vectors
    d_can = _canberra_dist(agg_A1, agg_A2)
    return d_can


def resistance_distance(
    A1, A2, p=2, renormalized=False, attributed=False, check_connected=True, beta=1
):
    """Compare two graphs using resistance distance (possibly renormalized).

    Parameters
    ----------
    A1, A2 : NumPy Matrices
        Adjacency matrices of graphs to be compared.

    p : float
        Function returns the p-norm of the flattened matrices.

    renormalized : Boolean, optional (default = False)
        If true, then renormalized resistance distance is computed.

    attributed : Boolean, optional (default=False)
        If true, then the resistance distance PER NODE is returned.

    check_connected : Boolean, optional (default=True)
        If false, then no check on connectivity is performed. See Notes of
        resistance_matrix for more information.

    beta : float, optional (default=1)
        A parameter used in the calculation of the renormalized resistance
        matrix. If using regular resistance, this is irrelevant.

    Returns:
    -------
    dist : float of numpy array
        The RR distance between the two graphs. If attributed is True, then
        vector distance per node is returned.

    Notes:
    -----
    The distance is calculated by assuming the nodes are in correspondence, and
    any nodes not present are treated as isolated by renormalized resistance.

    References:
    ----------

    See Also:
    --------
    resistance_matrix
    """
    # Calculate resistance matrices and compare
    if renormalized:
        # pad smaller adj. mat. so they're the same size
        n1, n2 = [A.shape[0] for A in [A1, A2]]
        N = max(n1, n2)
        A1, A2 = [_pad(A, N) for A in [A1, A2]]
        R1, R2 = [renormalized_res_mat(A, beta=beta) for A in [A1, A2]]
    else:
        R1, R2 = [
            resistance_matrix(A, check_connected=check_connected) for A in [A1, A2]
        ]
    try:
        distance_vector = np.sum((R1 - R2) ** p, axis=1)
    except ValueError:
        raise InputError(
            "Input matrices are different sizes. Please use "
            "renormalized resistance distance."
        )
    if attributed:
        return distance_vector ** (1 / p)

    return np.sum(distance_vector) ** (1 / p)


def deltacon0(A1, A2, eps=None):
    """DeltaCon0 distance between two graphs. The distance is the Frobenius norm
    of the element-wise square root of the fast belief propagation matrix.

    Parameters
    ----------
    A1, A2 : NumPy Matrices
        Adjacency matrices of graphs to be compared.

    Returns:
    -------
    dist : float
        DeltaCon0 distance between graphs.

    References:
    ----------

    See Also:
    --------
    fast_bp
    """
    # pad smaller adj. mat. so they're the same size
    n1, n2 = [A.shape[0] for A in [A1, A2]]
    N = max(n1, n2)
    A1, A2 = [_pad(A, N) for A in [A1, A2]]
    S1, S2 = [fast_bp(A, eps=eps) for A in [A1, A2]]
    dist = np.abs(np.sqrt(S1) - np.sqrt(S2)).sum()
    return dist



"""Graph utilities module for SWMManywhere.

A module to contain graphfcns, the graphfcn registry object, and other graph
utilities (such as save/load functions).
"""

from __future__ import annotations

import json
import tempfile
import warnings
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Callable, List, Optional

import networkx as nx
import shapely

from swmmanywhere import geospatial_utilities as go
from swmmanywhere import parameters
from swmmanywhere.filepaths import FilePaths
from swmmanywhere.logging import logger, verbose


def load_graph(fid: Path) -> nx.Graph:
    """Load a graph from a file saved with save_graph.

    Args:
        fid (Path): The path to the file

    Returns:
        G (nx.Graph): A graph
    """
    json_data = json.loads(fid.read_text())

    G = nx.node_link_graph(json_data, directed=True)
    for u, v, data in G.edges(data=True):
        if "geometry" in data:
            geometry_coords = data["geometry"]
            line_string = shapely.LineString(shapely.wkt.loads(geometry_coords))
            data["geometry"] = line_string
    return G


def _serialize_line_string(obj):
    if isinstance(obj, shapely.LineString):
        return obj.wkt
    return obj


def save_graph(G: nx.Graph, fid: Path) -> None:
    """Save a graph to a file.

    Args:
        G (nx.Graph): A graph
        fid (Path): The path to the file
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", FutureWarning)
        json_data = nx.node_link_data(G)

    with fid.open("w") as file:
        json.dump(json_data, file, default=_serialize_line_string)


class BaseGraphFunction(ABC):
    """Base class for graph functions.

    On a SWMManywhere project the intention is to iterate over a number of
    graph functions. Each graph function may require certain attributes to
    be present in the graph. Each graph function may add attributes to the
    graph. This class provides a framework for graph functions to check
    their requirements and additions a-priori when the list is provided.
    """

    required_edge_attributes: List[str] = []
    adds_edge_attributes: List[str] = []
    required_node_attributes: List[str] = []
    adds_node_attributes: List[str] = []

    def __init_subclass__(
        cls,
        required_edge_attributes: Optional[List[str]] = None,
        adds_edge_attributes: Optional[List[str]] = None,
        required_node_attributes: Optional[List[str]] = None,
        adds_node_attributes: Optional[List[str]] = None,
    ):
        """Set the required and added attributes for the subclass."""
        cls.required_edge_attributes = required_edge_attributes or []
        cls.adds_edge_attributes = adds_edge_attributes or []
        cls.required_node_attributes = required_node_attributes or []
        cls.adds_node_attributes = adds_node_attributes or []

    @abstractmethod
    def __call__(self, G: nx.Graph, *args, **kwargs) -> nx.Graph:
        """Run the graph function."""
        return G

    def validate_requirements(self, edge_attributes: set, node_attributes: set) -> None:
        """Validate that the graph has the required attributes."""
        for attribute in self.required_edge_attributes:
            assert attribute in edge_attributes, f"{attribute} not in edge attributes"

        for attribute in self.required_node_attributes:
            assert attribute in node_attributes, f"{attribute} not in node attributes"

    def add_graphfcn(
        self, edge_attributes: set, node_attributes: set
    ) -> tuple[set, set]:
        """Add the attributes that the graph function adds."""
        self.validate_requirements(edge_attributes, node_attributes)
        edge_attributes = edge_attributes.union(self.adds_edge_attributes)
        node_attributes = node_attributes.union(self.adds_node_attributes)
        return edge_attributes, node_attributes


class GraphFunctionRegistry(dict):
    """Registry object."""

    def register(self, cls):
        """Register a graph function."""
        if cls.__name__ in self:
            raise ValueError(f"{cls.__name__} already in the graph functions registry!")

        self[cls.__name__] = cls()
        return cls

    def __getattr__(self, name):
        """Get a graph function from the graphfcn dict."""
        try:
            return self[name]
        except KeyError:
            raise AttributeError(f"{name} NOT in the graph functions registry!")


graphfcns = GraphFunctionRegistry()


def register_graphfcn(cls) -> Callable:
    """Register a graph function.

    Args:
        cls (Callable): A class that inherits from BaseGraphFunction

    Returns:
        cls (Callable): The same class
    """
    graphfcns.register(cls)
    return cls


def filter_streets(G):
    """Filter streets.

    This function removes non streets from a graph.

    Args:
        G (nx.Graph): A graph

    Returns:
       (nx.Graph): A graph of only street edges
    """
    G = G.copy()
    # Remove non-street edges/nodes and unconnected nodes
    nodes_to_remove = []
    for u, v, d in G.edges(data=True):
        if d["edge_type"] != "street":
            if d["edge_type"] == "outfall":
                nodes_to_remove.append(v)
            else:
                nodes_to_remove.extend((u, v))
    G.remove_nodes_from(nodes_to_remove)
    return G


def validate_graphfcn_list(
    graphfcn_list: list[str], starting_graph: nx.Graph | None = None
) -> None:
    """Validate that the graph functions are registered.

    Tests that the graph functions are registered.

    Tests that the graph functions have the required attributes in the graph
    and updates the attributes that are added to the graph.
    `required_edge_attributes` and `required_node_attributes` currently only
    specify that one element in the graph must have the attribute (e.g., if a
    graphfcn has `required_node_attributes=['id']`, and only one node in the
    graph has the `id` attribute, then it will be valid).

    Args:
        graphfcn_list (list[str]): A list of graph functions
        starting_graph (nx.Graph, optional): A graph to check the starting
            attributes of. Defaults to None.

    Raises:
        ValueError: If a graph function is not registered
        ValueError: If a graph function requires an attribute that is not in
            the graph
    """
    # Check that the graph functions are registered
    not_exists = [g for g in graphfcn_list if g not in graphfcns]
    if not_exists:
        raise ValueError(f"Graphfcns are not registered:\n{', '.join(not_exists)}")

    if starting_graph is None:
        return

    # Get starting graph attributes
    edge_attributes: set = set()
    for u, v, data in starting_graph.edges(data=True):
        edge_attributes = edge_attributes.union(data.keys())

    node_attributes: set = set()
    for node, data in starting_graph.nodes(data=True):
        node_attributes = node_attributes.union(data.keys())

    # Iterate over graphfcn_list and check that the required attributes are
    # present in the graph, updating the add attributes
    for graphfcn in graphfcn_list:
        if node_attributes.intersection(
            graphfcns[graphfcn].required_node_attributes
        ) != set(graphfcns[graphfcn].required_node_attributes):
            raise ValueError(
                f"""Graphfcn {graphfcn} requires node attributes 
                {graphfcns[graphfcn].required_node_attributes}"""
            )
        if edge_attributes.intersection(
            graphfcns[graphfcn].required_edge_attributes
        ) != set(graphfcns[graphfcn].required_edge_attributes):
            raise ValueError(
                f"""Graphfcn {graphfcn} requires edge attributes 
                {graphfcns[graphfcn].required_edge_attributes}"""
            )

        edge_attributes = edge_attributes.union(
            graphfcns[graphfcn].adds_edge_attributes
        )
        node_attributes = node_attributes.union(
            graphfcns[graphfcn].adds_node_attributes
        )


with tempfile.TemporaryDirectory(dir=".") as temp_dir:
    temp_addresses = FilePaths(
        base_dir=Path(temp_dir), bbox_bounds=(0, 1, 0, 1), project_name="temp"
    )


def iterate_graphfcns(
    G: nx.Graph,
    graphfcn_list: list[str],
    params: dict = parameters.get_full_parameters(),
    addresses: FilePaths = temp_addresses,
) -> nx.Graph:
    """Iterate a list of graph functions over a graph.

    Args:
        G (nx.Graph): The graph to iterate over.
        graphfcn_list (list[str]): A list of graph functions to iterate.
        params (dict): A dictionary of parameters to pass to the graph
            functions.
        addresses (FilePaths): A FilePaths parameter object

    Returns:
        nx.Graph: The graph after the graph functions have been applied.
    """
    validate_graphfcn_list(graphfcn_list)

    for function in graphfcn_list:
        G = graphfcns[function](G, addresses=addresses, **params)
        if len(filter_streets(G).edges) == 0:
            logger.warning(
                f"""graphfcn: {function} removed all edges, 
                           returning graph."""
            )
            return G
        else:
            logger.info(f"graphfcn: {function} completed.")

        if verbose():
            save_graph(G, addresses.model_paths.model / f"{function}_graph.json")
            go.graph_to_geojson(
                graphfcns.fix_geometries(G),
                addresses.model_paths.model / f"{function}_nodes.geojson",
                addresses.model_paths.model / f"{function}_edges.geojson",
                G.graph["crs"],
            )
    return G



"""**************
Linear Algebra
**************

Linear algebraic functions, calculations of important matrices.
"""

from __future__ import annotations

from .eigenstuff import *
from .fast_bp import *
from .matrices import *

# import helper functions for use in other places
from .resistance import *



"""Test parameters."""

from __future__ import annotations

from pathlib import Path

from swmmanywhere import parameters
from swmmanywhere.swmmanywhere import import_module


def test_custom_parameters():
    """Test register_parameter_group."""
    import_module(Path(__file__).parent / "test_data" / "custom_parameters.py")

    assert "new_params" in parameters.get_full_parameters()
    assert "new_param" in parameters.get_full_parameters_flat()


def test_replace_parameter_group():
    """Test replacing a parameter group."""

    @parameters.register_parameter_group("topology_derivation")
    class NewTopologyDerivation(parameters.TopologyDerivation):
        new_weight: float = parameters.Field(
            default=1,
            le=1,
            ge=0,
        )

    # Check has new_weight
    assert hasattr(
        parameters.get_full_parameters()["topology_derivation"], "new_weight"
    )

    # Check retained existing weight
    assert hasattr(
        parameters.get_full_parameters()["topology_derivation"],
        "chahinian_angle_scaling",
    )



"""Utilities for YAML save/load.

Author: cheginit
"""

from __future__ import annotations

import base64
import functools
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Any

import folium
import geopandas as gpd
import pandas as pd
import yaml
from matplotlib import pyplot as plt

if TYPE_CHECKING:
    SafeDumper = yaml.SafeDumper
    from yaml.nodes import Node
else:
    Node = Any
    SafeDumper = getattr(yaml, "CSafeDumper", yaml.SafeDumper)

yaml_load = functools.partial(
    yaml.load, Loader=getattr(yaml, "CSafeLoader", yaml.SafeLoader)
)


class PathDumper(SafeDumper):
    """A dumper that can represent pathlib.Path objects as strings."""

    def represent_data(self, data: Any) -> Node:
        """Represent data."""
        if isinstance(data, Path):
            return self.represent_scalar("tag:yaml.org,2002:str", str(data))
        return super().represent_data(data)


def yaml_dump(o: Any, stream: Any = None, **kwargs: Any) -> str:
    """Dump YAML.

    Notes:
    -----
    When python/mypy#1484 is solved, this can be ``functools.partial``
    """
    return yaml.dump(
        o,
        Dumper=PathDumper,
        stream=stream,
        default_flow_style=False,
        indent=2,
        sort_keys=False,
        **kwargs,
    )


def plot_basic(nodes_path: Path, edges_path: Path):
    """Create a basic map with nodes and edges.

    Args:
        nodes_path (Path): The path to the nodes file.
        edges_path (Path): The path to the edges file.

    Returns:
        folium.Map: The folium map.
    """
    # Load and inspect results
    nodes = gpd.read_file(nodes_path)
    edges = gpd.read_file(edges_path)

    # Convert to EPSG 4326 for plotting
    nodes = nodes.to_crs(4326)
    edges = edges.to_crs(4326)

    # Identify outfalls
    if "outfall" in nodes.columns:
        outfall = nodes.id == nodes.outfall
    else:
        outfall = ~nodes.id.isin(edges.u.astype(str))

    # Plot on map
    m = folium.Map(
        location=[nodes.geometry.y.mean(), nodes.geometry.x.mean()], zoom_start=16
    )
    folium.GeoJson(edges, color="black", weight=1).add_to(m)
    folium.GeoJson(
        nodes.loc[~outfall],
        marker=folium.CircleMarker(
            radius=3,  # Radius in metres
            weight=0,  # outline weight
            fill_color="black",
            fill_opacity=1,
        ),
    ).add_to(m)

    folium.GeoJson(
        nodes.loc[outfall],
        marker=folium.CircleMarker(
            radius=3,  # Radius in metres
            weight=0,  # outline weight
            fill_color="red",
            fill_opacity=1,
        ),
    ).add_to(m)

    # Display the map
    return m


def plot_clickable(nodes_path: Path, edges_path: Path, results_path: Path):
    """Create a clickable map with nodes, edges and results.

    Args:
        nodes_path (Path): The path to the nodes file.
        edges_path (Path): The path to the edges file.
        results_path (Path): The path to the results file.

    Returns:
        folium.Map: The folium map.
    """
    # Load and inspect results
    nodes = gpd.read_file(nodes_path)
    edges = gpd.read_file(edges_path)
    df = pd.read_parquet(results_path)
    df.id = df.id.astype(str)
    floods = df.loc[df.variable == "flooding"].groupby("id")
    flows = df.loc[df.variable == "flow"].groupby("id")

    if "outfall" not in nodes.columns:
        nodes["outfall"] = None
        nodes.loc[~nodes.id.isin(edges.u.astype(str)), "outfall"] = nodes.id

    # Convert to EPSG 4326 for plotting
    nodes = nodes.to_crs(4326).set_index("id")
    edges = edges.to_crs(4326).set_index("id")

    # Create map
    m = folium.Map(
        location=[nodes.geometry.y.mean(), nodes.geometry.x.mean()], zoom_start=16
    )

    # Add edges
    for edge, row in edges.iterrows():
        # Create a plot for each edge
        grp = flows.get_group(str(edge))
        f, ax = plt.subplots(figsize=(4, 3))
        grp.set_index("date").value.plot(ylabel="flow (l/s)", title=edge, ax=ax)
        img = BytesIO()
        f.tight_layout()
        f.savefig(img, format="png", dpi=94)
        plt.close(f)

        # Convert plot to base64
        img.seek(0)
        img_base64 = base64.b64encode(img.read()).decode()
        img_html = f'<img src="data:image/png;base64,{img_base64}">'

        # Add edge to map
        folium.PolyLine(
            [[c[1], c[0]] for c in row.geometry.coords],
            color="black",
            weight=2,
            popup=folium.Popup(img_html),
        ).add_to(m)

    # Add nodes
    for node, row in nodes.iterrows():
        grp = floods.get_group(str(node))
        f, ax = plt.subplots(figsize=(4, 3))
        grp.set_index("date").value.plot(ylabel="flooding (l)", title=node, ax=ax)
        img = BytesIO()
        f.tight_layout()
        f.savefig(img, format="png", dpi=94)
        plt.close(f)
        img.seek(0)
        img_base64 = base64.b64encode(img.read()).decode()
        img_html = f'<img src="data:image/png;base64,{img_base64}">'
        if row.get("outfall") == node:
            color = "red"
        else:
            color = "black"
        folium.CircleMarker(
            [nodes.loc[node].geometry.y, nodes.loc[node].geometry.x],
            color=color,
            radius=3,
            weight=0,
            fill_color=color,
            fill_opacity=1,
            popup=folium.Popup(img_html),
        ).add_to(m)

    return m


def plot_map(model_dir: Path):
    """Create a map from a model directory.

    Args:
        model_dir (Path): The directory containing the model files.

    Returns:
        folium.Map: The folium map.
    """
    nodes_fids = list(model_dir.glob("nodes.*"))
    if not any(nodes_fids):
        raise FileNotFoundError("No nodes or edges found in model directory.")
    nodes = nodes_fids[0]

    edges_fids = list(model_dir.glob("edges.*"))
    if not any(edges_fids):
        raise FileNotFoundError("No edges found in model directory.")
    edges = edges_fids[0]

    results_fids = list(model_dir.glob("*results.*"))

    if results_fids:
        results = results_fids[0]
        m = plot_clickable(nodes, edges, results)
    else:
        m = plot_basic(nodes, edges)
    return m



"""Prepare data module for SWMManywhere.

A module to download data needed for SWMManywhere.
"""

from __future__ import annotations

import json
import operator
from contextlib import suppress
from datetime import datetime, timedelta
from pathlib import Path
from typing import cast

import cdsapi
import networkx as nx
import osmnx as ox
import pandas as pd
import planetary_computer
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.dataset as ds
import pyarrow.parquet as pq
import pystac_client
import requests
import rioxarray
import rioxarray.merge as rxr_merge
import xarray as xr
from geopy.geocoders import Nominatim
from packaging.version import Version
from pyarrow import fs

from swmmanywhere.logging import logger
from swmmanywhere.utilities import yaml_load


def get_country(x: float, y: float) -> dict[int, str]:
    """Identify which country a point (x,y) is in.

    Return the two and three letter ISO code for coordinates x (longitude)
    and y (latitude)

    Args:
        x (float): Longitude.
        y (float): Latitude.

    Returns:
        dict: A dictionary containing the two and three letter ISO codes.
            Example: {2: 'GB', 3: 'GBR'}
    """
    # Get the directory of the current module
    current_dir = Path(__file__).parent

    # TODO use 'load_yaml_from_defs'
    # Create the path to iso_converter.yml
    iso_path = current_dir / "defs" / "iso_converter.yml"

    # Initialize geolocator
    geolocator = Nominatim(user_agent="get_iso")

    # Load ISO code mapping from YAML file
    data = yaml_load(iso_path.read_text())

    # Get country ISO code from coordinates
    location = geolocator.reverse(f"{y}, {x}", exactly_one=True)
    iso_country_code = location.raw.get("address", {}).get("country_code").upper()

    # Return a dictionary with the two and three letter ISO codes
    return {2: iso_country_code, 3: data.get(iso_country_code, "")}


def _get_latest_s3_url() -> tuple[str, str]:
    """Get S3 URL with caching based on Overture's monthly release policy.

    Returns:
    -------
    s3_region : str
        The S3 region
    s3_url : str
        The S3 url with the latest overturmaps' version
    """
    cache_file = Path(".cache/overture_release.json")
    cache_file.parent.mkdir(exist_ok=True)
    s3_region = "us-west-2"

    # Check cache (valid for 72 hours given monthly releases)
    if cache_file.exists():
        cache = json.loads(cache_file.read_text())
        cached_time = datetime.fromisoformat(cache["timestamp"])

        # If cache is less than 72 hours old, use it
        if datetime.now() - cached_time < timedelta(hours=72):
            url = (
                f"overturemaps-us-west-2/release/{cache['release']}/theme=buildings"
                "/type=building/"
            )
            return s3_region, url

        # If we have the release date and it's been less than 28 days, keep using it
        with suppress(Exception):
            release_date = datetime.strptime(cache["release"].split(".")[0], "%Y-%m-%d")
            if (datetime.now() - release_date).days < 28:
                url = (
                    f"overturemaps-us-west-2/release/{cache['release']}/theme=buildings"
                    "/type=building/"
                )
                return s3_region, url

    response = requests.get("https://stac.overturemaps.org/catalog.json")
    catalog = response.json()

    releases = [
        link
        for link in catalog["links"]
        if link["rel"] == "child" and "release" in link.get("title", "").lower()
    ]
    releases.sort(key=operator.itemgetter("href"), reverse=True)
    latest = str(Path(releases[0]["href"].rstrip("/")).parent)
    cache_file.write_text(
        json.dumps({"release": latest, "timestamp": datetime.now().isoformat()})
    )
    url = f"overturemaps-{s3_region}/release/{latest}/theme=buildings/type=building/"
    return s3_region, url


def download_buildings_bbox(
    file_address: Path, bbox: tuple[float, float, float, float]
) -> None:
    """Retrieve building data in bbox from Overture Maps to file.

    This function is based on
    `overturemaps-py <https://github.com/OvertureMaps/overturemaps-py>`__.

    Args:
        bbox (tuple): Bounding box coordinates (xmin, ymin, xmax, ymax)
        file_address (Path): File address to save the downloaded data.
    """
    s3_region, s3_url = _get_latest_s3_url()
    xmin, ymin, xmax, ymax = bbox
    filter = (
        (pc.field("bbox", "xmin") < xmax)
        & (pc.field("bbox", "xmax") > xmin)
        & (pc.field("bbox", "ymin") < ymax)
        & (pc.field("bbox", "ymax") > ymin)
    )

    dataset = ds.dataset(
        s3_url, filesystem=fs.S3FileSystem(anonymous=True, region=s3_region)
    )
    batches = dataset.to_batches(filter=filter)
    non_empty_batches = (b for b in batches if b.num_rows > 0)

    geoarrow_schema = dataset.schema.set(
        dataset.schema.get_field_index("geometry"),
        dataset.schema.field("geometry").with_metadata(
            {b"ARROW:extension:name": b"geoarrow.wkb"}
        ),
    )
    reader = pa.RecordBatchReader.from_batches(geoarrow_schema, non_empty_batches)

    with pq.ParquetWriter(file_address, reader.schema) as writer:
        for batch in reader:
            if batch.num_rows > 0:
                writer.write_batch(batch)


def download_buildings(file_address: Path, x: float, y: float) -> int:
    """Download buildings data based on coordinates and save to a file.

    Args:
        file_address (Path): File address to save the downloaded data.
        x (float): Longitude.
        y (float): Latitude.

    Returns:
        status_code (int): Response status code
    """
    # Get three letter ISO code
    iso_country_code = get_country(x, y)[3]

    # Construct API URL
    api_url = f"https://data.source.coop/vida/google-microsoft-open-buildings/geoparquet/by_country/country_iso={iso_country_code}/{iso_country_code}.parquet"

    # Download data
    response = requests.get(api_url)
    if response.status_code == 200:
        # Save data to the specified file address
        with file_address.open("wb") as file:
            file.write(response.content)
        logger.info(f"Data downloaded and saved to {file_address}")
    else:
        logger.error(f"Error downloading data. Status code: {response.status_code}")
    return response.status_code


def download_street(
    bbox: tuple[float, float, float, float], network_type="drive"
) -> nx.MultiDiGraph:
    """Get street network within a bounding box using OSMNX.

    [CREDIT: Taher Cheghini busn_estimator package]

    Args:
        bbox (tuple[float, float, float, float]): Bounding box as tuple in form
            of (west, south, east, north) at EPSG:4326.
        network_type (str, optional): Type of street network. Defaults to 'drive'.

    Returns:
        nx.MultiDiGraph: Street network with type drive and
            ``truncate_by_edge set`` to True.
    """
    if Version(ox.__version__) < Version("2.0.0b1"):
        west, south, east, north = bbox
        bbox = (north, south, east, west)
    graph = ox.graph_from_bbox(
        bbox=bbox, network_type=network_type, truncate_by_edge=True
    )
    return cast("nx.MultiDiGraph", graph)


def download_river(bbox: tuple[float, float, float, float]) -> nx.MultiDiGraph:
    """Get water network within a bounding box using OSMNX.

    Args:
        bbox (tuple[float, float, float, float]): Bounding box as tuple in form
            of (west, south, east, north) at EPSG:4326.

    Returns:
        nx.MultiDiGraph: River network with type waterway and
            ``truncate_by_edge set`` to True.
    """
    try:
        if Version(ox.__version__) < Version("2.0.0b1"):
            west, south, east, north = bbox
            bbox = (north, south, east, west)
        graph = ox.graph_from_bbox(
            bbox=bbox,
            truncate_by_edge=True,
            custom_filter='["waterway"]',
            retain_all=True,
        )
    except ValueError as e:
        if "Found no graph nodes within the requested polygon" in str(e):
            logger.warning("No water network found within the bounding box.")
            return nx.MultiDiGraph()
        else:
            raise  # Re-raise the exception

    return cast("nx.MultiDiGraph", graph)


def download_elevation(fid: Path, bbox: tuple[float, float, float, float]) -> None:
    """Download NASADEM elevation data from Microsoft Planetary computer.

    Downloads elevation data in GeoTIFF format from Microsoft Planetary computer
      based on the specified bounding box.

    Args:
        fid (Path): File path to save the downloaded elevation data.
        bbox (tuple[float, float, float, float]): Bounding box as tuple in form
            of (west, south, east, north) at EPSG:4326.

    Example:
        ```
        bbox = (-120, 35, -118, 37)  # Example bounding box coordinates
        download_elevation('elevation_data.tif',
                            bbox)
        ```

    Author:
        cheginit
    """
    catalog = pystac_client.Client.open(
        "https://planetarycomputer.microsoft.com/api/stac/v1",
        modifier=planetary_computer.sign_inplace,
    )
    search = catalog.search(
        collections=["nasadem"],
        bbox=bbox,
    )
    signed_asset = (
        planetary_computer.sign(item.assets["elevation"]).href
        for item in search.items()
    )
    dem = rxr_merge.merge_arrays(
        [rioxarray.open_rasterio(href).squeeze(drop=True) for href in signed_asset]
    )
    dem = dem.rio.clip_box(*bbox)
    dem.rio.to_raster(fid)


def download_precipitation(
    bbox: tuple[float, float, float, float],
    start_date: str = "2015-01-01",
    end_date: str = "2015-01-05",
    username: str = "<your_username>",
    api_key: str = "<your_api_key>",
) -> pd.DataFrame:
    """Download precipitation data within bbox from ERA5.

    Register for an account and API key at: https://cds.climate.copernicus.eu.
    Produces hourly gridded data (31km grid) in CRS EPSG:4326.
    More information at: https://github.com/ecmwf/cdsapi.

    Args:
        bbox (tuple): Bounding box coordinates in the format
            (minx, miny, maxx, maxy).
        start_date (str, optional): Start date. Defaults to '2015-01-01'.
        end_date (str, optional): End date. Defaults to '2015-01-05'.
        username (str, optional): CDS api username.
            Defaults to '<your_username>'.
        api_key (str, optional): CDS api key. Defaults to '<your_api_key>'.

    Returns:
        df (DataFrame): DataFrame containing downloaded data.
    """
    # Notes for docstring:
    # looks like api will give nearest point if not in bbox

    # Define the request parameters
    request = {
        "product_type": "reanalysis",
        "format": "netcdf",
        "variable": "total_precipitation",
        "date": "/".join([start_date, end_date]),
        "time": "/".join([f"{str(x).zfill(2)}:00" for x in range(24)]),
        "area": bbox,
    }

    c = cdsapi.Client(
        url="https://cds.climate.copernicus.eu/api/v2", key=f"{username}:{api_key}"
    )
    # Get data
    c.retrieve("reanalysis-era5-single-levels", request, "download.nc")

    with xr.open_dataset("download.nc") as data:
        # Convert the xarray Dataset to a pandas DataFrame
        df = data.to_dataframe()
        df["unit"] = "m/hr"

    # Delete nc file
    Path("download.nc").unlink()

    return df



"""Geospatial utilities module for SWMManywhere.

A module containing functions to perform a variety of geospatial operations,
such as reprojecting coordinates and handling raster data.
"""

from __future__ import annotations

import itertools
import json
import math
import os
import shutil
import tempfile
from functools import lru_cache
from pathlib import Path
from typing import List, Optional

import geopandas as gpd
import networkx as nx
import numpy as np
import pandas as pd
import pyproj
import rasterio as rst
import rioxarray
import shapely
from pywbt import prepare_wbt, whitebox_tools
from rasterio import features
from scipy.interpolate import RegularGridInterpolator
from scipy.spatial import KDTree
from shapely import geometry as sgeom
from shapely.strtree import STRtree
from tqdm.auto import tqdm

from swmmanywhere.logging import logger, verbose

os.environ["NUMBA_NUM_THREADS"] = "1"
import pyflwdir  # noqa: E402

TransformerFromCRS = lru_cache(pyproj.transformer.Transformer.from_crs)


def get_utm_epsg(
    x: float,
    y: float,
    crs: str | int | pyproj.CRS = "EPSG:4326",
    datum_name: str = "WGS 84",
) -> str:
    """Get the UTM CRS code for a given coordinate.

    Note, this function is taken from GeoPandas and modified to use
    for getting the UTM CRS code for a given coordinate.

    Args:
        x (float): Longitude in crs
        y (float): Latitude in crs
        crs (str | int | pyproj.CRS, optional): The CRS of the input
            coordinates. Defaults to 'EPSG:4326'.
        datum_name (str, optional): The datum name to use for the UTM CRS

    Returns:
        str: Formatted EPSG code for the UTM zone.

    Example:
        >>> get_utm_epsg(-0.1276, 51.5074)
        'EPSG:32630'
    """
    if not isinstance(x, float) or not isinstance(y, float):
        raise TypeError("x and y must be floats")

    try:
        crs = pyproj.CRS(crs)
    except pyproj.exceptions.CRSError:
        raise ValueError("Invalid CRS")

    # ensure using geographic coordinates
    if pyproj.CRS(crs).is_geographic:
        lon = x
        lat = y
    else:
        transformer = TransformerFromCRS(crs, "EPSG:4326", always_xy=True)
        lon, lat = transformer.transform(x, y)
    utm_crs_list = pyproj.database.query_utm_crs_info(
        datum_name=datum_name,
        area_of_interest=pyproj.aoi.AreaOfInterest(
            west_lon_degree=lon,
            south_lat_degree=lat,
            east_lon_degree=lon,
            north_lat_degree=lat,
        ),
    )
    return f"{utm_crs_list[0].auth_name}:{utm_crs_list[0].code}"


def interp_with_nans(
    xy: tuple[float, float],
    interp: RegularGridInterpolator,
    grid: np.ndarray,
    values: list[float],
) -> float:
    """Wrap the interpolation function to handle NaNs.

    Picks the nearest non NaN grid point if the interpolated value is NaN,
    otherwise returns the interpolated value.

    Args:
        xy (tuple): Coordinate of interest
        interp (RegularGridInterpolator): The interpolator object.
        grid (np.ndarray): List of xy coordinates of the grid points.
        values (list): The list of values at each point in the grid.

    Returns:
        float: The interpolated value.
    """
    # Call the interpolator
    val = float(interp(xy))
    # If the value is NaN, we need to pick nearest non nan grid point
    if np.isnan(val):
        # Get the distances to all grid points
        distances = np.linalg.norm(grid - xy, axis=1)
        # Get the indices of the grid points sorted by distance
        indices = np.argsort(distances)
        # Iterate over the grid points in order of increasing distance
        for index in indices:
            # If the value at this grid point is not NaN, return it
            if not np.isnan(values[index]):
                return values[index]
    else:
        return val

    raise ValueError("No non NaN values found in grid.")


def interpolate_points_on_raster(
    x: list[float], y: list[float], elevation_fid: Path
) -> list[float]:
    """Interpolate points on a raster.

    Args:
        x (list): X coordinates.
        y (list): Y coordinates.
        elevation_fid (Path): Filepath to elevation raster.

    Returns:
        elevation (float): Elevation at point.
    """
    with rst.open(elevation_fid) as src:
        # Read the raster data
        data = src.read(1).astype(float)  # Assuming it's a single-band raster
        data[data == src.nodata] = None

        # Get the raster's coordinates
        x_grid = np.linspace(src.bounds.left, src.bounds.right, src.width)
        y_grid = np.linspace(src.bounds.bottom, src.bounds.top, src.height)

        # Define grid
        xx, yy = np.meshgrid(x_grid, y_grid)
        grid = np.vstack([xx.ravel(), yy.ravel()]).T
        values = data.ravel()

        # Define interpolator
        interp = RegularGridInterpolator(
            (y_grid, x_grid),
            np.flipud(data),
            method="linear",
            bounds_error=False,
            fill_value=None,
        )
        # Interpolate for x,y
        return [
            interp_with_nans((y_, x_), interp, grid, values) for x_, y_ in zip(x, y)
        ]


def reproject_raster(
    target_crs: str, fid: Path, new_fid: Optional[Path] = None
) -> None:
    """Reproject a raster to a new CRS.

    Args:
        target_crs (str): Target CRS in EPSG format (e.g., EPSG:32630).
        fid (Path): Filepath to the raster to reproject.
        new_fid (Path, optional): Filepath to save the reprojected raster.
            Defaults to None, which will just use fid with '_reprojected'.
    """
    # Open the raster
    with rioxarray.open_rasterio(fid) as raster:
        # Reproject the raster
        reprojected = raster.rio.reproject(target_crs)

        # Define the output filepath
        if new_fid is None:
            new_fid = Path(str(fid.with_suffix("")) + "_reprojected.tif")

        # Save the reprojected raster
        reprojected.rio.to_raster(new_fid)


def get_transformer(source_crs: str, target_crs: str) -> pyproj.Transformer:
    """Get a transformer object for reprojection.

    Args:
        source_crs (str): Source CRS in EPSG format (e.g., EPSG:32630).
        target_crs (str): Target CRS in EPSG format (e.g., EPSG:32630).

    Returns:
        pyproj.Transformer: Transformer object for reprojection.

    Example:
        >>> transformer = get_transformer('EPSG:4326', 'EPSG:32630')
        >>> x, y = transformer.transform(-0.1276, 51.5074)
        >>> print(f"{x:.6f}, {y:.6f}")
        699330.110690, 5710164.303007
    """
    return pyproj.Transformer.from_crs(source_crs, target_crs, always_xy=True)


def reproject_df(df: pd.DataFrame, source_crs: str, target_crs: str) -> pd.DataFrame:
    """Reproject the coordinates in a DataFrame.

    Args:
        df (pd.DataFrame): DataFrame with columns 'longitude' and 'latitude'.
        source_crs (str): Source CRS in EPSG format (e.g., EPSG:4326).
        target_crs (str): Target CRS in EPSG format (e.g., EPSG:32630).
    """
    # Function to transform coordinates
    pts = gpd.points_from_xy(df["longitude"], df["latitude"], crs=source_crs).to_crs(
        target_crs
    )
    df = df.copy()
    df["x"] = pts.x
    df["y"] = pts.y
    return df


def reproject_graph(G: nx.Graph, source_crs: str, target_crs: str) -> nx.Graph:
    """Reproject the coordinates in a graph.

    osmnx.projection.project_graph might be suitable if some other behaviour
    needs to be captured, but it currently fails the tests so I will ignore for
    now.

    Args:
        G (nx.Graph): Graph with nodes containing 'x' and 'y' properties.
        source_crs (str): Source CRS in EPSG format (e.g., EPSG:4326).
        target_crs (str): Target CRS in EPSG format (e.g., EPSG:32630).

    Returns:
        nx.Graph: Graph with nodes containing 'x' and 'y' properties.
    """
    # Create a PyProj transformer for CRS conversion
    transformer = get_transformer(source_crs, target_crs)

    # Create a new graph with the converted nodes and edges
    G_new = G.copy()

    # Convert and add nodes with 'x', 'y' properties
    for node, data in G_new.nodes(data=True):
        x, y = transformer.transform(data["x"], data["y"])
        data["x"] = x
        data["y"] = y

    # Convert and add edges with 'geometry' property
    for u, v, data in G_new.edges(data=True):
        if "geometry" in data.keys():
            data["geometry"] = sgeom.LineString(
                itertools.starmap(transformer.transform, data["geometry"].coords)
            )
        else:
            data["geometry"] = sgeom.LineString(
                [
                    [G_new.nodes[u]["x"], G_new.nodes[u]["y"]],
                    [G_new.nodes[v]["x"], G_new.nodes[v]["y"]],
                ]
            )
    G_new.graph["crs"] = target_crs
    return G_new


def nearest_node_buffer(
    points1: dict[str, sgeom.Point], points2: dict[str, sgeom.Point], threshold: float
) -> dict:
    """Find the nearest node within a given buffer threshold.

    Args:
        points1 (dict): A dictionary where keys are labels and values are
            Shapely points geometries.
        points2 (dict): A dictionary where keys are labels and values are
            Shapely points geometries.
        threshold (float): The maximum distance for a node to be considered
            'nearest'. If no nodes are within this distance, the node is not
            included in the output.

    Returns:
        dict: A dictionary where keys are labels from points1 and values are
            labels from points2 of the nearest nodes within the threshold.
    """
    if not points1 or not points2:
        return {}

    # Convert the keys of points2 to a list
    labels2 = list(points2.keys())

    # Create a spatial index
    tree = STRtree(list(points2.values()))

    # Initialize an empty dictionary to store the matching nodes
    matching = {}

    # Iterate over points1
    for key, geom in points1.items():
        # Find the nearest node in the spatial index to the current geometry
        nearest = tree.nearest(geom)
        nearest_geom = points2[labels2[nearest]]

        # If the nearest node is within the threshold, add it to the
        # matching dictionary
        if geom.buffer(threshold).intersects(nearest_geom):
            matching[key] = labels2[nearest]

    # Return the matching dictionary
    return matching


def burn_shape_in_raster(
    geoms: list[sgeom.LineString], depth: float, raster_fid: Path, new_raster_fid: Path
) -> None:
    """Burn a depth into a raster along a list of shapely geometries.

    Args:
        geoms (list): List of Shapely geometries.
        depth (float): Depth to carve.
        raster_fid (Path): Filepath to input raster.
        new_raster_fid (Path): Filepath to save the carved raster.
    """
    with rst.open(raster_fid) as src:
        # read data
        data = src.read(1)
        data = data.astype(float)
        data_mask = data != src.nodata
        bool_mask = np.zeros(data.shape, dtype=bool)
        for geom in geoms:
            # Create a mask for the line
            mask = features.geometry_mask(
                [sgeom.mapping(geom)],
                out_shape=src.shape,
                transform=src.transform,
                invert=True,
            )
            # modify masked data
            bool_mask[mask] = True  # Adjust this multiplier as needed
        # modify data
        data[bool_mask & data_mask] -= depth
        # Create a new GeoTIFF with modified values
        with rst.open(
            new_raster_fid,
            "w",
            driver="GTiff",
            height=src.height,
            width=src.width,
            count=1,
            dtype=data.dtype,
            crs=src.crs,
            transform=src.transform,
            nodata=src.nodata,
        ) as dest:
            dest.write(data, 1)


def remove_intersections(polys: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """Remove intersections from a GeoDataFrame of polygons.

    Subcatchments are derived for a given point, and so larger subcatchments
    will contain smaller ones. This function removes the smaller subcatchments
    from the larger ones.

    Args:
        polys (gpd.GeoDataFrame): A GeoDataFrame containing polygons with
            columns: 'geometry', 'area', and 'id'.

    Returns:
        gpd.GeoDataFrame: A GeoDataFrame containing polygons with no
            intersections.
    """
    result_polygons = polys.copy()

    # Sort the polygons by area (smallest first)
    result_polygons = (
        result_polygons.assign(area=result_polygons.geometry.area)
        .sort_values("area", ascending=True)
        .reset_index(drop=True)
    )

    # Store the area of 'trimmed' polygons into a combined geometry, starting
    # with the smallest area polygon
    minimal_geom = result_polygons.iloc[0]["geometry"]
    for idx, row in tqdm(
        result_polygons.iloc[1:].iterrows(),
        total=result_polygons.shape[0] - 1,
        disable=not verbose(),
    ):
        # Trim the polygon by the combined geometry
        result_polygons.at[idx, "geometry"] = row["geometry"].difference(minimal_geom)

        # Update the combined geometry with the new polygon
        minimal_geom = minimal_geom.union(row["geometry"])

    # Sort the polygons by area (largest first) - this is just to conform to
    # the unit test and is not strictly essential
    result_polygons = result_polygons.sort_values("area", ascending=False).drop(
        "area", axis=1
    )

    return result_polygons


def remove_zero_area_subareas(
    mp: sgeom.MultiPolygon, removed_subareas: List[sgeom.Polygon]
) -> sgeom.MultiPolygon:
    """Remove subareas with zero area from a multipolygon.

    Args:
        mp (sgeom.MultiPolygon): A multipolygon.
        removed_subareas (List[sgeom.Polygon]): A list of removed subareas.

    Returns:
        sgeom.MultiPolygon: A multipolygon with zero area subareas removed.
    """
    if not hasattr(mp, "geoms"):
        return mp

    largest = max(mp.geoms, key=lambda x: x.area)
    removed = [subarea for subarea in mp.geoms if subarea != largest]
    removed_subareas.extend(removed)
    return largest


def attach_unconnected_subareas(
    polys_gdf: gpd.GeoDataFrame, unconnected_subareas: List[sgeom.Polygon]
) -> gpd.GeoDataFrame:
    """Attach unconnected subareas to the nearest polygon.

    Args:
        polys_gdf (gpd.GeoDataFrame): A GeoDataFrame containing polygons with
            columns: 'geometry', 'area', and 'id'.
        unconnected_subareas (List[sgeom.Polygon]): A list of subareas that are
            not attached to others.

    Returns:
        gpd.GeoDataFrame: A GeoDataFrame containing polygons.
    """
    tree = STRtree(polys_gdf.geometry)
    for subarea in unconnected_subareas:
        nearest_poly = tree.nearest(subarea)
        ind = polys_gdf.index[nearest_poly]
        new_poly = polys_gdf.loc[ind, "geometry"].union(subarea)
        if hasattr(new_poly, "geoms"):
            new_poly = max(new_poly.geoms, key=lambda x: x.area)
        polys_gdf.at[ind, "geometry"] = new_poly
    return polys_gdf


def calculate_slope(
    polys_gdf: gpd.GeoDataFrame, grid: Grid, cell_slopes: np.ndarray
) -> gpd.GeoDataFrame:
    """Calculate the average slope of each polygon.

    Args:
        polys_gdf (gpd.GeoDataFrame): A GeoDataFrame containing polygons with
            columns: 'geometry', 'area', and 'id'.
        grid (Grid): Information of the raster (affine, shape, crs, bbox)
        cell_slopes (np.ndarray): The slopes of each cell in the grid.

    Returns:
        gpd.GeoDataFrame: A GeoDataFrame containing polygons with an added
            'slope' column.
    """
    polys_gdf["slope"] = None
    for idx, row in polys_gdf.iterrows():
        mask = features.geometry_mask(
            [row.geometry], grid.shape, grid.affine, invert=True
        )
        average_slope = cell_slopes[mask].mean().mean()
        polys_gdf.loc[idx, "slope"] = max(float(average_slope), 0)
    return polys_gdf


def vectorize(
    data: np.ndarray,
    nodata: float,
    transform: rst.Affine,
    crs: int,
    name: str = "value",
) -> gpd.GeoDataFrame:
    """Vectorize raster data into a geodataframe.

    Args:
        data (np.ndarray): The raster data.
        nodata (float): The nodata value.
        transform (rst.Affine): The affine transformation.
        crs (int): The CRS of the data.
        name (str, optional): The name of the data. Defaults to "value".

    Returns:
        gpd.GeoDataFrame: A GeoDataFrame containing the vectorized data.
    """
    feats_gen = features.shapes(
        data,
        mask=data != nodata,
        transform=transform,
        connectivity=8,
    )
    feats = [
        {"geometry": geom, "properties": {name: val}} for geom, val in list(feats_gen)
    ]

    # parse to geopandas for plotting / writing to file
    gdf = gpd.GeoDataFrame.from_features(feats, crs=crs)
    gdf[name] = gdf[name].astype(data.dtype)
    return gdf


def delineate_catchment_pyflwdir(
    grid: Grid, flow_dir: np.array, G: nx.Graph
) -> gpd.GeoDataFrame:
    """Derive subcatchments from the nodes on a graph and a DEM.

    Uses the pyflwdir catchment delineation functionality. About a magnitude
    faster than delineate_catchment.

    Args:
        grid (Grid): Information of the raster (affine, shape, crs, bbox).
        flow_dir (np.array): Flow directions.
        G (nx.Graph): The input graph with nodes containing 'x' and 'y'.

    Returns:
        gpd.GeoDataFrame: A GeoDataFrame containing polygons with columns:
            'geometry', 'area', 'id', 'width', and 'slope'.
    """
    flw = pyflwdir.from_array(
        flow_dir,
        ftype="d8",
        check_ftype=False,
        transform=grid.affine,
    )
    bbox = sgeom.box(*grid.bbox)
    u, x, y = zip(
        *[
            (u, float(p["x"]), float(p["y"]))
            for u, p in G.nodes(data=True)
            if sgeom.Point(p["x"], p["y"]).within(bbox)
        ]
    )

    subbasins = flw.basins(xy=(x, y))
    gdf_bas = vectorize(
        subbasins.astype(np.int32), 0, flw.transform, G.graph["crs"], name="basin"
    )
    gdf_bas["id"] = [u[x - 1] for x in gdf_bas["basin"]]
    return gdf_bas


def derive_subbasins_streamorder(
    fid: Path,
    streamorder: int | None = None,
    x: list[float] = [],
    y: list[float] = [],
    wbt_zip_path: Path | None = None,
) -> gpd.GeoDataFrame:
    """Derive subbasins.

    Use the pyflwdir snap function to find the most downstream points in each
    subbasin. If streamorder is provided it will use that instead, although
    defaulting to snap if there are no cells of the correct streamorder.

    Args:
        fid (Path): Filepath to the DEM.
        streamorder (int): The stream order to delineate subbasins for.
        x (list): X coordinates.
        y (list): Y coordinates.
        wbt_zip_path (Path, optional): Path to WhiteboxTools binaries. Defaults to None.

    Returns:
        gpd.GeoDataFrame: A GeoDataFrame containing polygons.
    """
    # Load and process the DEM
    grid, flow_dir, _ = load_and_process_dem(fid, wbt_zip_path=wbt_zip_path)

    flw = pyflwdir.from_array(
        flow_dir,
        ftype="d8",
        check_ftype=False,
        transform=grid.affine,
    )
    xy = [
        (x_, y_)
        for x_, y_ in zip(x, y)
        if (x_ > grid.bbox[0])
        and (x_ < grid.bbox[2])
        and (y_ > grid.bbox[1])
        and (y_ < grid.bbox[3])
    ]

    idxs, _ = flw.snap(xy=list(zip(*xy)))
    subbasins = flw.basins(idxs=np.unique(idxs))

    if streamorder is not None:
        # Identify stream order
        subbasins_, _ = flw.subbasins_streamorder(min_sto=streamorder)

        if np.unique(subbasins_).shape[0] == 1:
            logger.warning(
                """No subbasins found in `derive_subbasins_streamorder`. 
                    Instead subbasins have been selected based on the most downstream 
                    points. But you should inspect `subbasins` and probably check your 
                    DEM."""
            )
        else:
            subbasins = subbasins_

    gdf_bas = vectorize(
        subbasins.astype(np.int32), 0, flw.transform, grid.crs, name="basin"
    )

    return gdf_bas


def flwdir_whitebox(fid: Path, wbt_zip_path: Path | None = None) -> np.array:
    """Calculate flow direction using WhiteboxTools.

    Args:
        fid (Path): Filepath to the DEM.
        wbt_zip_path (Path, optional): Path to WhiteboxTools binaries. If the binaries
            are not present, they will be downloaded. If None, the binaries will be
            stored temporarily and removed after the function call. Defaults to None.


    Returns:
        np.array: Flow directions.
    """
    # Initialize WhiteboxTools
    with tempfile.TemporaryDirectory(dir=str(fid.parent)) as temp_dir:
        temp_path = Path(temp_dir)

        # Copy raster to working directory
        dem = temp_path / "dem.tif"
        shutil.copy(fid, dem)

        wbt_root = temp_path / "WBT"

        prepare_wbt(wbt_root, zip_path=wbt_zip_path)

        # Condition
        wbt_args = {
            "BreachDepressions": ["-i=dem.tif", "--fillpits", "-o=dem_corr.tif"],
            "D8Pointer": ["-i=dem_corr.tif", "-o=fdir.tif"],
        }

        whitebox_tools(
            temp_path,
            wbt_args,
            save_dir=temp_path,
            verbose=verbose(),
            wbt_root=wbt_root,
            zip_path=wbt_zip_path,
            max_procs=1,
        )

        shutil.rmtree(wbt_root)

        fdir = temp_path / "fdir.tif"
        if not Path(fdir).exists():
            raise ValueError("Flow direction raster not created.")

        with rst.open(fdir) as src:
            flow_dir = src.read(1)

    # Adjust mapping from WhiteboxTools to pyflwdir
    mapping = {1: 128, 2: 1, 4: 2, 8: 4, 16: 8, 32: 16, 64: 32, 128: 64}
    get_flow_dir = np.vectorize(mapping.get, excluded=["default"])
    flow_dir = get_flow_dir(flow_dir, 0)
    return flow_dir


class Grid:
    """A class to represent a grid."""

    def __init__(self, affine: rst.Affine, shape: tuple, crs: int, bbox: tuple):
        """Initialize the Grid class.

        Args:
            affine (rst.Affine): The affine transformation.
            shape (tuple): The shape of the grid.
            crs (int): The CRS of the grid.
            bbox (tuple): The bounding box of the grid.
        """
        self.affine = affine
        self.shape = shape
        self.crs = crs
        self.bbox = bbox


def load_and_process_dem(
    fid: Path,
    method: str = "whitebox",
    wbt_zip_path: Path | None = None,
) -> tuple[Grid, np.array, np.array]:
    """Load and condition a DEM.

    Args:
        fid (Path): Filepath to the DEM.
        method (str, optional): The method to use for conditioning. Defaults to
            "whitebox".
        wbt_zip_path (Path, optional): Path to WhiteboxTools binaries. Defaults to None.

    Returns:
        tuple: A tuple containing the grid, flow directions, and cell slopes.
    """
    with rst.open(fid, "r") as src:
        elevtn = src.read(1).astype(float)
        nodata = float(src.nodata)
        transform = src.transform
        crs = src.crs

    if method not in ("whitebox", "pyflwdir"):
        raise ValueError("Method must be 'whitebox' or 'pyflwdir'.")

    if method == "whitebox":
        flow_dir = flwdir_whitebox(fid, wbt_zip_path=wbt_zip_path)
    elif method == "pyflwdir":
        flw = pyflwdir.from_dem(
            data=elevtn,
            nodata=nodata,
            transform=transform,
            latlon=crs.is_geographic,
        )
        flow_dir = flw.to_array(ftype="d8").astype(int)

    cell_slopes = pyflwdir.dem.slope(
        elevtn,
        nodata=nodata,
        transform=transform,
        latlon=crs.is_geographic,
    )

    grid = Grid(transform, elevtn.shape, crs, src.bounds)

    return grid, flow_dir, cell_slopes


def derive_subcatchments(
    G: nx.Graph, fid: Path, method: str = "whitebox", wbt_zip_path: Path | None = None
) -> gpd.GeoDataFrame:
    """Derive subcatchments from the nodes on a graph and a DEM.

    Args:
        G (nx.Graph): The input graph with nodes containing 'x' and 'y'.
        fid (Path): Filepath to the DEM.
        method (str, optional): The method to use for conditioning.
        wbt_zip_path (Path, optional): Path to WhiteboxTools binaries. Defaults to None.

    Returns:
        gpd.GeoDataFrame: A GeoDataFrame containing polygons with columns:
            'geometry', 'area', 'id', 'width', and 'slope'.
    """
    # Load and process the DEM
    grid, flow_dir, cell_slopes = load_and_process_dem(
        fid, method, wbt_zip_path=wbt_zip_path
    )

    # Delineate catchments
    result_polygons = delineate_catchment_pyflwdir(grid, flow_dir, G)

    # Convert to GeoDataFrame
    polys_gdf = result_polygons.dropna(subset=["geometry"])
    polys_gdf = polys_gdf[~polys_gdf["geometry"].is_empty]

    # Remove zero area subareas and attach to nearest polygon
    removed_subareas: List[sgeom.Polygon] = []  # needed for mypy

    def remove_(mp):
        return remove_zero_area_subareas(mp, removed_subareas)

    polys_gdf["geometry"] = polys_gdf["geometry"].apply(remove_)
    polys_gdf = attach_unconnected_subareas(polys_gdf, removed_subareas)

    # Calculate area and slope
    polys_gdf["area"] = polys_gdf.geometry.area
    polys_gdf = calculate_slope(polys_gdf, grid, cell_slopes)

    # Calculate width
    polys_gdf["width"] = polys_gdf["area"].div(np.pi).pow(0.5)
    return polys_gdf


def derive_rc(
    subcatchments: gpd.GeoDataFrame,
    building_footprints: gpd.GeoDataFrame,
    streetcover: gpd.GeoDataFrame,
) -> gpd.GeoDataFrame:
    """Derive the Runoff Coefficient (RC) of each subcatchment.

    The runoff coefficient is the ratio of impervious area to total area. The
    impervious area is calculated by overlaying the subcatchments with building
    footprints and all edges in G buffered by their width parameter (i.e., to
    calculate road area).

    Args:
        subcatchments (gpd.GeoDataFrame): A GeoDataFrame containing polygons that
            represent subcatchments with columns: 'geometry', 'area', and 'id'.
        building_footprints (gpd.GeoDataFrame): A GeoDataFrame containing
            building footprints with a 'geometry' column.
        streetcover (gpd.GeoDataFrame): A GeoDataFrame containing street cover
            with a 'geometry' column.

    Returns:
        gpd.GeoDataFrame: A GeoDataFrame containing polygons with columns:
            'geometry', 'area', 'id', 'impervious_area', and 'rc'.

    Author:
        @cheginit, @barneydobson
    """
    # Map buffered streets and buildings to subcatchments
    subcat_tree = subcatchments.sindex
    impervious = gpd.GeoDataFrame(
        pd.concat([building_footprints[["geometry"]], streetcover[["geometry"]]]),
        crs=building_footprints.crs,
    )
    bf_pidx, sb_pidx = subcat_tree.query(impervious.geometry, predicate="intersects")
    sb_idx = subcatchments.iloc[sb_pidx].index

    # Calculate impervious area and runoff coefficient (rc)
    subcatchments["impervious_area"] = 0.0

    # Calculate all intersection-impervious geometries
    intersection_area = shapely.intersection(
        subcatchments.iloc[sb_pidx].geometry.to_numpy(),
        impervious.iloc[bf_pidx].geometry.to_numpy(),
    )

    # Indicate which catchment each intersection is part of
    intersections = pd.DataFrame(
        [
            {"sb_idx": ix, "impervious_geometry": ia}
            for ix, ia in zip(sb_idx, intersection_area)
        ]
    )

    # Aggregate by catchment
    areas = (
        intersections.groupby("sb_idx")
        .apply(shapely.ops.unary_union)
        .apply(shapely.area)
    )

    # Store as impervious area in subcatchments
    subcatchments["impervious_area"] = 0.0
    subcatchments.loc[areas.index, "impervious_area"] = areas
    subcatchments["rc"] = (
        subcatchments["impervious_area"] / subcatchments.geometry.area * 100
    )
    return subcatchments


def calculate_angle(
    point1: tuple[float, float],
    point2: tuple[float, float],
    point3: tuple[float, float],
) -> float:
    """Calculate the angle between three points.

    Calculate the angle between the vectors formed by (point1,
    point2) and (point2, point3)

    Args:
        point1 (tuple): The first point (x,y).
        point2 (tuple): The second point (x,y).
        point3 (tuple): The third point (x,y).

    Returns:
        float: The angle between the three points in degrees.
    """
    vector1 = (point1[0] - point2[0], point1[1] - point2[1])
    vector2 = (point3[0] - point2[0], point3[1] - point2[1])

    dot_product = vector1[0] * vector2[0] + vector1[1] * vector2[1]
    magnitude1 = math.sqrt(vector1[0] ** 2 + vector1[1] ** 2)
    magnitude2 = math.sqrt(vector2[0] ** 2 + vector2[1] ** 2)

    if magnitude1 * magnitude2 == 0:
        # Avoid division by zero
        return float("inf")

    cosine_angle = dot_product / (magnitude1 * magnitude2)

    # Ensure the cosine value is within the valid range [-1, 1]
    cosine_angle = min(max(cosine_angle, -1), 1)

    # Calculate the angle in radians
    angle_radians = math.acos(cosine_angle)

    # Convert angle to degrees
    angle_degrees = math.degrees(angle_radians)

    return angle_degrees


def nodes_to_features(G: nx.Graph):
    """Convert a graph to a GeoJSON node feature collection.

    Args:
        G (nx.Graph): The input graph.

    Returns:
        dict: A GeoJSON feature collection.
    """
    features = []
    for node, data in G.nodes(data=True):
        feature = {
            "type": "Feature",
            "geometry": sgeom.mapping(sgeom.Point(data["x"], data["y"])),
            "properties": {"id": node, **data},
        }
        features.append(feature)
    return features


def edges_to_features(G: nx.Graph):
    """Convert a graph to a GeoJSON edge feature collection.

    Args:
        G (nx.Graph): The input graph.

    Returns:
        dict: A GeoJSON feature collection.
    """
    features = []
    for u, v, data in G.edges(data=True):
        if "geometry" not in data:
            geom = None
        else:
            geom = sgeom.mapping(data["geometry"])
            del data["geometry"]
        feature = {
            "type": "Feature",
            "geometry": geom,
            "properties": {"u": u, "v": v, **data},
        }
        features.append(feature)
    return features


def graph_to_geojson(graph: nx.Graph, fid_nodes: Path, fid_edges: Path, crs: str):
    """Write a graph to a GeoJSON file.

    Args:
        graph (nx.Graph): The input graph.
        fid_nodes (Path): The filepath to save the nodes GeoJSON file.
        fid_edges (Path): The filepath to save the edges GeoJSON file.
        crs (str): The CRS of the graph.
    """
    graph = graph.copy()
    nodes = nodes_to_features(graph)
    edges = edges_to_features(graph)

    for iterable, fid in zip([nodes, edges], [fid_nodes, fid_edges]):
        geojson = {
            "type": "FeatureCollection",
            "features": iterable,
            "crs": {
                "type": "name",
                "properties": {
                    "name": "urn:ogc:def:crs:{0}".format(crs.replace(":", "::"))
                },
            },
        }

        with fid.open("w") as output_file:
            json.dump(geojson, output_file, indent=2)


def merge_points(coordinates: list[tuple[float, float]], threshold: float) -> dict:
    """Merge points that are within a threshold distance.

    Args:
        coordinates (list): List of coordinates as tuples.
        threshold(float): The threshold distance for merging points.

    Returns:
        dict: A dictionary mapping the original point index to the merged point
            and new coordinate.
    """
    # Create a KDTtree to pair together points within thresholds
    tree = KDTree(coordinates)
    pairs = tree.query_pairs(threshold)

    # Merge pairs into families of points that are all nearby
    families: list = []

    for pair in pairs:
        matched_families = [
            family for family in families if pair[0] in family or pair[1] in family
        ]

        if matched_families:
            # Merge all matched families and add the current pair
            new_family = set(pair)
            for family in matched_families:
                new_family.update(family)

            # Remove the old families and add the newly formed one
            for family in matched_families:
                families.remove(family)
            families.append(new_family)
        else:
            # No matching family found, so create a new one
            families.append(set(pair))

    # Create a mapping of the original point to the merged point
    mapping = {}
    for family in families:
        average_point = np.mean([coordinates[i] for i in family], axis=0)
        family_head = min(list(family))
        for i in family:
            mapping[i] = {"maps_to": family_head, "coordinate": tuple(average_point)}
    return mapping



from __future__ import annotations

import networkx as nx
import pytest

from swmmanywhere.shortest_path_utils import dijkstra_pq, tarjans_pq


def test_simple_graph():
    """Test case 1: Simple connected graph."""
    G = nx.MultiDiGraph()
    G.add_edges_from([(2, 1), (3, 1), (4, 2), (4, 3), (1, "waste")])
    nx.set_edge_attributes(
        G,
        {
            (2, 1, 0): {"weight": 1, "edge_type": "outfall"},
            (3, 1, 0): {"weight": 2, "edge_type": "street"},
            (4, 2, 0): {"weight": 3, "edge_type": "street"},
            (4, 3, 0): {"weight": 4, "edge_type": "street"},
            (1, "waste", 0): {"weight": 0, "edge_type": "outfall"},
        },
    )
    root = "waste"
    mst = tarjans_pq(G, root)
    assert set(mst.edges) == {(1, "waste", 0), (2, 1, 0), (4, 2, 0), (3, 1, 0)}
    assert mst.size() == 4

    djg = dijkstra_pq(G, [root])
    assert set(djg.edges) == {(1, "waste", 0), (2, 1, 0), (4, 2, 0), (3, 1, 0)}
    assert djg.size() == 4


def test_disconnected():
    """Test case 2: Disconnected graph."""
    G = nx.MultiDiGraph()
    G.add_edges_from(
        [(1, "waste"), (2, 1), (3, 1), (5, 4), (6, 5)], weight=1, edge_type="street"
    )
    root = "waste"
    with pytest.raises(ValueError):
        tarjans_pq(G, root)

    djg = dijkstra_pq(G, [root])
    assert set(djg.edges) == {(1, "waste", 0), (2, 1, 0), (3, 1, 0)}


def test_parallel():
    """Test case 3: Graph with parallel edges."""
    G = nx.MultiDiGraph()
    G.add_edges_from(
        [(2, 1, 0), (2, 1, 1), (3, 1, 0), (4, 2, 0), (4, 3, 0)],
        edge_type="street",
        weight=1,
    )
    root = 1
    mst = tarjans_pq(G, root)
    assert set(mst.edges) == {(2, 1, 0), (4, 2, 0), (3, 1, 0)}
    assert mst.size() == 3

    djg = dijkstra_pq(G, [root])
    # Currently paths are defined as node-to-node and so ignore keys .. TODO?
    assert set(djg.edges) == {(2, 1, 0), (2, 1, 1), (3, 1, 0), (4, 2, 0)}


def test_selfloop():
    """Test case 4: Graph with self-loops."""
    G = nx.MultiDiGraph()
    G.add_edges_from(
        [(2, 1, 0), (3, 1, 0), (4, 2, 0), (4, 3, 0), (2, 4, 0)],
        edge_type="street",
        weight=1,
    )
    G.add_edge(3, 4, weight=1, edge_type="street")
    root = 1
    mst = tarjans_pq(G, root)
    assert set(mst.edges) == {(2, 1, 0), (4, 2, 0), (3, 1, 0)}
    assert mst.size() == 3

    djg = dijkstra_pq(G, [root])
    assert set(djg.edges) == {(2, 1, 0), (4, 2, 0), (3, 1, 0)}
    assert djg.size() == 3


def test_custom_weight():
    """Test case 5: Graph with custom weight attribute."""
    G = nx.MultiDiGraph()
    G.add_edges_from(
        [(2, 1, 0), (3, 1, 0), (4, 2, 0), (4, 3, 0)], edge_type="street", cost=1
    )
    root = 1
    mst = tarjans_pq(G, root, weight_attr="cost")
    assert set(mst.edges) == {(2, 1, 0), (4, 2, 0), (3, 1, 0)}
    assert mst.size() == 3

    djg = dijkstra_pq(G, [root], weight_attr="cost")
    assert set(djg.edges) == {(2, 1, 0), (4, 2, 0), (3, 1, 0)}
    assert djg.size() == 3


def test_empty():
    """Test case 6: Empty graph."""
    G = nx.MultiDiGraph()
    root = 1
    with pytest.raises(KeyError):
        tarjans_pq(G, root)
    with pytest.raises(nx.NetworkXError):
        dijkstra_pq(G, [root])



"""Debug results by recalculating metrics.

This script provides a way to load a model file from the default setup in
experimenter.py and recalculate the metrics. This is useful for recreating
how a metric is calculated to verify that it is being done correctly. In this
example we reproduce code from `metric_utilities.py` to check how timeseries
data are aligned and compared.
"""

from __future__ import annotations

from pathlib import Path

import geopandas as gpd
import pandas as pd

from swmmanywhere.graph_utilities import load_graph
from swmmanywhere.metric_utilities import (
    align_by_shape,
    best_outfall_match,
    dominant_outfall,
    extract_var,
    iterate_metrics,
)
from swmmanywhere.parameters import MetricEvaluation
from swmmanywhere.swmmanywhere import load_config

if __name__ == "main":
    project = "cranbrook"
    base = Path.home() / "Documents" / "data" / "swmmanywhere"
    config_path = base / project / f"{project}_hpc.yml"
    config = load_config(config_path, validation=False)
    config["base_dir"] = base / project
    real_dir = config["base_dir"] / "real"

    model_number = 5523

    model_dir = config["base_dir"] / "bbox_1" / f"model_{model_number}"

    syn_results = pd.read_parquet(model_dir / "results.parquet")
    real_results = pd.read_parquet(real_dir / "real_results.parquet")

    syn_G = load_graph(model_dir / "assign_id_graph.json")
    real_G = load_graph(real_dir / "graph.json")

    syn_subcatchments = gpd.read_file(model_dir / "subcatchments.geoparquet")
    real_subcatchments = gpd.read_file(real_dir / "subcatchments.geojson")

    syn_metrics = iterate_metrics(
        syn_results,
        syn_subcatchments,
        syn_G,
        real_results,
        real_subcatchments,
        real_G,
        ["grid_nse_flooding", "subcatchment_nse_flooding"],
        MetricEvaluation(),
    )

    # Check outfall scale
    synthetic_results = syn_results.copy()
    real_results_ = real_results.copy()
    sg_syn, syn_outfall = best_outfall_match(syn_G, real_subcatchments)
    sg_real, real_outfall = dominant_outfall(real_G, real_results)

    # Check nnodes
    print(f"n syn nodes {len(sg_syn.nodes)}")
    print(f"n real nodes {len(sg_real.nodes)}")

    # Check contributing area
    # syn_subcatchments['impervious_area'].sum() / syn_subcatchments['area'].sum()
    # real_subcatchments['impervious_area'].sum() / real_subcatchments['area'].sum()
    variable = "flooding"

    # e.g., subs
    results = align_by_shape(
        variable,
        synthetic_results=synthetic_results,
        real_results=real_results,
        shapes=real_subcatchments,
        synthetic_G=syn_G,
        real_G=real_G,
    )

    # e.g., outfall
    if variable == "flow":
        syn_ids = [d["id"] for u, v, d in syn_G.edges(data=True) if v == syn_outfall]
        real_ids = [d["id"] for u, v, d in real_G.edges(data=True) if v == real_outfall]
    else:
        syn_ids = list(sg_syn.nodes)
        real_ids = list(sg_real.nodes)
    synthetic_results["date"] = pd.to_datetime(synthetic_results["date"])
    real_results["date"] = pd.to_datetime(real_results["date"])

    # Help alignment
    synthetic_results["id"] = synthetic_results["id"].astype(str)
    real_results["id"] = real_results["id"].astype(str)
    syn_ids = [str(x) for x in syn_ids]
    real_ids = [str(x) for x in real_ids]
    # Extract data
    syn_data = extract_var(synthetic_results, variable)
    syn_data = syn_data.loc[syn_data["id"].isin(syn_ids)]
    syn_data = syn_data.groupby("date").value.sum()

    real_data = extract_var(real_results, variable)
    real_data = real_data.loc[real_data["id"].isin(real_ids)]
    real_data = real_data.groupby("date").value.sum()

    # Align data
    df = pd.merge(
        syn_data,
        real_data,
        left_index=True,
        right_index=True,
        suffixes=("_syn", "_real"),
        how="outer",
    ).sort_index()

    # Interpolate to time in real data
    df["value_syn"] = df.value_syn.interpolate().to_numpy()
    df = df.dropna(subset=["value_real"])



"""********
Features
********

Calculation of features for NetSimile algorithm.
"""

from __future__ import annotations

import networkx as nx
import numpy as np
from scipy import stats

from ..linalg.matrices import _eps  # type: ignore


def get_features(A):
    """Feature grabber for NetSimile algorithm. Features used are

        1. Degree of node
        2. Clustering coefficient of node
        3. Average degree of node's neighbors
        4. Average clustering coefficient of node's neighbors
        5. Number of edges in node's egonet
        6. Number of neighbors of node's egonet
        7. Number of outgoing edges from node's egonet

    Parameters
    ---------
    A : NumPy matrix
        Adjacency matrix of graph in question. Preferably a SciPy sparse matrix
        for large graphs.

    Returns:
    -------
    feature_mat : NumPy array
        An n by 7 array of features, where n = A.shape[0]

    References:
    -----
    [Berlingerio 2012]

    """
    try:
        G = nx.from_scipy_sparse_array(A)
    except AttributeError:
        G = nx.from_numpy_array(A)
    n = len(G)
    # degrees, array so we can slice nice
    d_vec = np.array(list(dict(G.degree()).values()))
    # list of clustering coefficient
    clust_vec = np.array(list(nx.clustering(G).values()))
    neighbors = [G.neighbors(i) for i in range(n)]
    # average degree of neighbors (0 if node is isolated)
    neighbors = [list(G.neighbors(i)) for i in range(n)]
    # average degree of neighbors (0 if node is isolated)
    neighbor_deg = [
        d_vec[neighbors[i]].sum() / d_vec[i] if d_vec[i] > _eps else 0 for i in range(n)
    ]
    # avg. clustering coefficient of neighbors (0 if node is isolated)
    neighbor_clust = [
        clust_vec[neighbors[i]].sum() / d_vec[i] if d_vec[i] > _eps else 0
        for i in range(n)
    ]
    egonets = [nx.ego_graph(G, i) for i in range(n)]
    # number of edges in egonet
    ego_size = [G.number_of_edges() for G in egonets]
    # number of neighbors of egonet
    ego_neighbors = [
        len(
            set.union(*[set(neighbors[j]) for j in egonets[i].nodes()])
            - set(egonets[i].nodes())
        )
        for i in range(n)
    ]
    # number of edges outgoing from egonet
    outgoing_edges = [
        len(
            [
                edge
                for edge in G.edges(egonets[i].nodes())
                if edge[1] not in egonets[i].nodes()
            ]
        )
        for i in range(n)
    ]
    # use mat.T so that each node is a row (standard format)
    feature_mat = np.array(
        [
            d_vec,
            clust_vec,
            neighbor_deg,
            neighbor_clust,
            ego_size,
            ego_neighbors,
            outgoing_edges,
        ]
    ).T
    return feature_mat


def aggregate_features(feature_mat, row_var=False, as_matrix=False):
    """Returns column-wise descriptive statistics of a feature matrix.

    Parameters
    ----------
    feature_mat : NumPy array
        Matrix on which statistics are to be calculated. Assumed to be formatted
        so each row is an observation (a node, in the case of NetSimile).

    row_var : Boolean, optional (default=False)
        If true, then each variable has it's own row, and statistics are
        computed along rows rather than columns.

    as_matrix : Boolean, optional (default=False)
        If true, then description is returned as matrix. Otherwise, it is
        flattened into a vector.

    Returns:
    -------
    description : NumPy array
        Descriptive statistics of feature_mat

    Notes:
    -----

    References:
    ----------
    """
    axis = int(row_var)  # 0 if column-oriented, 1 if not
    description = np.array(
        [
            feature_mat.mean(axis=axis),
            np.median(feature_mat, axis=axis),
            np.std(feature_mat, axis=axis),
            stats.skew(feature_mat, axis=axis),
            stats.kurtosis(feature_mat, axis=axis),
        ]
    )
    if not as_matrix:
        description = description.flatten()
    return description



"""***********************
Fast Belief Propagation
***********************

The fast approximation of the Belief propagation matrix.
"""

from __future__ import annotations

import numpy as np
from numpy import linalg as la
from scipy import sparse as sps


def fast_bp(A, eps=None):
    """Return the fast belief propagation matrix of graph associated with A.

    Parameters
    ----------
    A : NumPy matrix or Scipy sparse matrix
        Adjacency matrix of a graph. If sparse, can be any format; CSC or CSR
        recommended.

    eps : float, optional (default=None)
        Small parameter used in calculation of matrix. If not provided, it is
        set to 1/(1+d_max) where d_max is the maximum degree.

    Returns:
    -------
    S : NumPy matrix or Scipy sparse matrix
        The fast belief propagation matrix. If input is sparse, will be returned
        as (sparse) CSC matrix.

    Notes:
    -----

    References:
    ----------

    """
    n, m = A.shape
    ##
    ## TODO: implement checks on the adjacency matrix
    ##
    degs = np.array(A.sum(axis=1)).flatten()
    if eps is None:
        eps = 1 / (1 + max(degs))
    identity = sps.identity(n)
    D = sps.dia_matrix((degs, [0]), shape=(n, n))
    # form inverse of S and invert (slow!)
    Sinv = identity + eps**2 * D - eps * A
    try:
        S = la.inv(Sinv)
    except Exception:
        Sinv = sps.csc_matrix(Sinv)
        S = sps.linalg.inv(Sinv)
    return S



"""Logging module for SWMManywhere.

Example:
```python
>>> import os
>>> os.environ["SWMMANYWHERE_VERBOSE"] = "true"
>>> # logging is now enabled in any swmmanywhere module
>>> from swmmanywhere.logging import logger # You can now log yourself
>>> logger.add("file.log") # Add a log file
>>> logger.info("This is an info message.") # Write to stdout and file.log
This is an info message.
>>> os.environ["SWMMANYWHERE_VERBOSE"] = "false" # Disable logging
```
"""

from __future__ import annotations

import os
import sys

import loguru


def verbose() -> bool:
    """Get the verbosity."""
    return os.getenv("SWMMANYWHERE_VERBOSE", "false").lower() == "true"


def set_verbose(verbose: bool):
    """Set the verbosity."""
    os.environ["SWMMANYWHERE_VERBOSE"] = str(verbose).lower()


def dynamic_filter(record):
    """A dynamic filter."""
    return verbose()


def get_logger() -> loguru.logger:
    """Get a logger."""
    logger = loguru.logger
    logger.configure(
        handlers=[
            {
                "sink": sys.stdout,
                "filter": dynamic_filter,
                "colorize": True,
                "format": " | ".join(
                    [
                        "<cyan>{time:YYYY/MM/DD HH:mm:ss}</>",
                        "{message}",
                    ]
                ),
            }
        ]
    )
    return logger


# Get the logger
logger = get_logger()

# Add a test_logger method to the logger
logger.test_logger = lambda: logger.info("This is a test message.")

# Store the original add method
original_add = logger.add


# Define a new function that wraps the original add method
def new_add(sink, **kwargs):
    """A new add method to wrap existing one but with the filter."""
    # Include the dynamic filter in the kwargs if not already specified
    if "filter" not in kwargs:
        kwargs["filter"] = dynamic_filter
    # Call the original add method with the updated kwargs
    return original_add(sink, **kwargs)


# Replace the logger's add method with new_add
logger.add = new_add



"""Misc line to trigger workflow."""

from __future__ import annotations

import tempfile
from pathlib import Path
from unittest.mock import MagicMock, patch

import geopandas as gpd
import networkx as nx
import numpy as np
import pytest
import rasterio as rst
from scipy.interpolate import RegularGridInterpolator
from shapely import geometry as sgeom

from swmmanywhere import geospatial_utilities as go
from swmmanywhere import graph_utilities as ge
from swmmanywhere.logging import set_verbose
from swmmanywhere.misc.debug_derive_rc import derive_rc_alt

test_data_dir = Path(__file__).parent / "test_data"


@pytest.fixture
def street_network():
    """Load a street network."""
    G = ge.load_graph(test_data_dir / "street_graph.json")
    return G


def almost_equal(a, b, tol=1e-6):
    """Check if two numbers are almost equal."""
    if hasattr(a, "shape"):
        return ((a - b) < tol).all().all()
    return abs(a - b) < tol


def test_interp_with_nans():
    """Test the interp_interp_with_nans function."""
    # Define a simple grid and values
    x = np.linspace(0, 1, 5)
    y = np.linspace(0, 1, 5)
    xx, yy = np.meshgrid(x, y)
    grid = np.vstack([xx.ravel(), yy.ravel()]).T
    values = np.linspace(0, 1, 25)
    values_grid = values.reshape(5, 5)

    # Define an interpolator
    interp = RegularGridInterpolator((x, y), values_grid)

    # Test the function at a point inside the grid
    yx = (0.875, 0.875)
    result = go.interp_with_nans(yx, interp, grid, values)
    assert result == 0.875

    # Test the function on a nan point
    values_grid[1][1] = np.nan
    yx = (0.251, 0.25)
    result = go.interp_with_nans(yx, interp, grid, values)
    assert result == values_grid[1][2]


@patch("rasterio.open")
def test_interpolate_points_on_raster(mock_rst_open):
    """Test the interpolate_points_on_raster function."""
    # Mock the raster file
    mock_src = MagicMock()
    mock_src.read.return_value = np.array([[1, 2], [3, 4]])
    mock_src.bounds = MagicMock()
    mock_src.bounds.left = 0
    mock_src.bounds.right = 1
    mock_src.bounds.bottom = 0
    mock_src.bounds.top = 1
    mock_src.width = 2
    mock_src.height = 2
    mock_src.nodata = None
    mock_rst_open.return_value.__enter__.return_value = mock_src

    # Define the x and y coordinates
    x = [0.25, 0.75]
    y = [0.25, 0.75]

    # Call the function
    result = go.interpolate_points_on_raster(x, y, Path("fake_path"))

    # [2.75, 2.25] feels unintuitive but it's because rasters measure from the top
    assert result == [2.75, 2.25]


def test_get_utm():
    """Test the get_utm_epsg function."""
    # Test a northern hemisphere point
    crs = go.get_utm_epsg(-1.0, 51.0)
    assert crs == "EPSG:32630"

    # Test a southern hemisphere point
    crs = go.get_utm_epsg(-1.0, -51.0)
    assert crs == "EPSG:32730"


def create_raster(fid):
    """Define a function to create a mock raster file."""
    data = np.ones((100, 100))
    transform = rst.transform.from_origin(0, 0, 0.1, 0.1)
    with rst.open(
        fid,
        "w",
        driver="GTiff",
        height=100,
        width=100,
        count=1,
        dtype="uint8",
        crs="EPSG:4326",
        transform=transform,
    ) as src:
        src.write(data, 1)


def test_reproject_raster():
    """Test the reproject_raster function."""
    # Create a mock raster file
    fid = Path("test.tif")
    try:
        create_raster(fid)

        # Define the input parameters
        target_crs = "EPSG:32630"
        new_fid = Path("test_reprojected.tif")

        # Call the function
        go.reproject_raster(target_crs, fid)

        # Check if the reprojected file exists
        assert new_fid.exists()

        # Check if the reprojected file has the correct CRS
        with rst.open(new_fid) as src:
            assert src.crs.to_string() == target_crs
    finally:
        # Regardless of test outcome, delete the temp file
        fid.unlink(missing_ok=True)
        new_fid.unlink(missing_ok=True)


def test_get_transformer():
    """Test the get_transformer function."""
    # Test a northern hemisphere point
    transformer = go.get_transformer("EPSG:4326", "EPSG:32630")

    initial_point = (-0.1276, 51.5074)
    expected_point = (699330.1106898375, 5710164.30300683)
    new_point = transformer.transform(*initial_point)
    assert almost_equal(new_point[0], expected_point[0])
    assert almost_equal(new_point[1], expected_point[1])


def test_reproject_graph():
    """Test the reproject_graph function."""
    # Create a mock graph
    G = nx.Graph()
    G.add_node(1, x=0, y=0)
    G.add_node(2, x=1, y=1)
    G.add_edge(1, 2)
    G.add_node(3, x=1, y=2)
    G.add_edge(2, 3, geometry=sgeom.LineString([(1, 1), (1, 2)]))

    # Define the input parameters
    source_crs = "EPSG:4326"
    target_crs = "EPSG:32630"

    # Call the function
    G_new = go.reproject_graph(G, source_crs, target_crs)

    # Test node coordinates
    assert almost_equal(G_new.nodes[1]["x"], 833978.5569194595)
    assert almost_equal(G_new.nodes[1]["y"], 0)
    assert almost_equal(G_new.nodes[2]["x"], 945396.6839773951)
    assert almost_equal(G_new.nodes[2]["y"], 110801.83254625657)
    assert almost_equal(G_new.nodes[3]["x"], 945193.8596723974)
    assert almost_equal(G_new.nodes[3]["y"], 221604.0105092727)

    # Test edge geometry
    assert almost_equal(list(G_new[1][2]["geometry"].coords)[0][0], 833978.5569194595)
    assert almost_equal(list(G_new[2][3]["geometry"].coords)[0][0], 945396.6839773951)


def test_nearest_node_buffer():
    """Test the nearest_node_buffer function."""
    # Create mock dictionaries of points
    points1 = {"a": sgeom.Point(0, 0), "b": sgeom.Point(1, 1)}
    points2 = {"c": sgeom.Point(0.5, 0.5), "d": sgeom.Point(2, 2)}

    # Define the input threshold
    threshold = 1.0

    # Call the function
    matching = go.nearest_node_buffer(points1, points2, threshold)

    # Check if the function returns the correct matching nodes
    assert matching == {"a": "c", "b": "c"}


def test_burn_shape_in_raster():
    """Test the burn_shape_in_raster function."""
    # Create a mock geometry
    geoms = [
        sgeom.LineString([(0, 0), (1, 1)]),
        sgeom.Polygon([(0, 0), (1, 0), (1, 1), (0, 1)]),
    ]

    # Define the input parameters
    depth = 1.0
    raster_fid = Path("input.tif")
    new_raster_fid = Path("output.tif")
    try:
        create_raster(raster_fid)

        # Call the function
        go.burn_shape_in_raster(geoms, depth, raster_fid, new_raster_fid)

        with rst.open(raster_fid) as src:
            data_ = src.read(1)

        # Open the new GeoTIFF file and check if it has been correctly modified
        with rst.open(new_raster_fid) as src:
            data = src.read(1)
            assert (data != data_).any()
    finally:
        # Regardless of test outcome, delete the temp file
        raster_fid.unlink(missing_ok=True)
        new_raster_fid.unlink(missing_ok=True)


@pytest.mark.parametrize(
    "method,area,slope,width",
    [("pyflwdir", 2498, 0.1187, 28.202), ("whitebox", 2998, 0.1102, 30.894)],
)
@pytest.mark.parametrize("verbose", [True, False])
def test_derive_subcatchments(
    wbt_zip_path, street_network, method, area, slope, width, verbose
):
    """Test the derive_subcatchments function."""
    set_verbose(verbose)

    elev_fid = test_data_dir / "elevation.tif"
    with patch(
        "swmmanywhere.geospatial_utilities.flwdir_whitebox", wraps=go.flwdir_whitebox
    ) as spy:
        polys = go.derive_subcatchments(
            street_network, elev_fid, method=method, wbt_zip_path=wbt_zip_path
        )
        assert all(
            kwargs["wbt_zip_path"] is not None for _, kwargs in spy.call_args_list
        )

    assert "slope" in polys.columns
    assert "area" in polys.columns
    assert "geometry" in polys.columns
    assert "id" in polys.columns
    assert polys.shape[0] > 0
    assert polys.dropna().shape == polys.shape
    assert polys.crs == street_network.graph["crs"]

    assert almost_equal(polys.set_index("id").loc[2623975694, "area"], area, tol=1)
    assert almost_equal(
        polys.set_index("id").loc[2623975694, "slope"], slope, tol=0.001
    )
    assert almost_equal(
        polys.set_index("id").loc[2623975694, "width"], width, tol=0.001
    )


def test_derive_rc(street_network):
    """Test the derive_rc function."""
    crs = street_network.graph["crs"]
    eg_bldg = sgeom.Polygon(
        [
            (700291, 5709928),
            (700331, 5709927),
            (700321, 5709896),
            (700293, 5709900),
            (700291, 5709928),
        ]
    )
    buildings = gpd.GeoDataFrame(geometry=[eg_bldg], crs=crs)
    subs = [
        sgeom.Polygon(
            [
                (700262, 5709928),
                (700262, 5709883),
                (700351, 5709883),
                (700351, 5709906),
                (700306, 5709906),
                (700306, 5709928),
                (700262, 5709928),
            ]
        ),
        sgeom.Polygon(
            [
                (700306, 5709928),
                (700284, 5709928),
                (700284, 5709950),
                (700374, 5709950),
                (700374, 5709906),
                (700351, 5709906),
                (700306, 5709906),
                (700306, 5709928),
            ]
        ),
        sgeom.Polygon(
            [
                (700351, 5709883),
                (700351, 5709906),
                (700374, 5709906),
                (700374, 5709883),
                (700396, 5709883),
                (700396, 5709816),
                (700329, 5709816),
                (700329, 5709838),
                (700329, 5709883),
                (700351, 5709883),
            ]
        ),
    ]

    streetcover = [
        d["geometry"].buffer(5) for u, v, d in street_network.edges(data=True)
    ]
    streetcover = gpd.GeoDataFrame(geometry=streetcover, crs=crs)

    subs = gpd.GeoDataFrame(
        data={"id": [107733, 1696030874, 6277683849]}, geometry=subs, crs=crs
    )
    subs["area"] = subs.geometry.area

    # Test no RC
    subs_rc = go.derive_rc(subs, buildings, buildings).set_index("id")
    assert subs_rc.loc[6277683849, "impervious_area"] == 0
    assert subs_rc.loc[107733, "impervious_area"] > 0

    # Test alt method
    subs_rc_alt = derive_rc_alt(subs, buildings, buildings).set_index("id")
    assert almost_equal(
        subs_rc[["area", "impervious_area", "rc"]],
        subs_rc_alt[["area", "impervious_area", "rc"]],
    )

    # Test some RC
    subs_rc = go.derive_rc(subs, buildings, streetcover).set_index("id")
    assert almost_equal(subs_rc.loc[6277683849, "impervious_area"], 1092.452579)
    assert almost_equal(subs_rc.loc[6277683849, "rc"], 21.770677)
    assert subs_rc.rc.max() <= 100

    # Test alt method
    subs_rc_alt = derive_rc_alt(subs, buildings, streetcover).set_index("id")
    assert almost_equal(
        subs_rc[["area", "impervious_area", "rc"]],
        subs_rc_alt[["area", "impervious_area", "rc"]],
    )

    # Test intersecting buildings and streets
    buildings = buildings.overlay(streetcover.dissolve(), how="union")
    subs_rc = go.derive_rc(subs, buildings, streetcover).set_index("id")
    assert almost_equal(subs_rc.loc[6277683849, "impervious_area"], 1092.452579)
    assert almost_equal(subs_rc.loc[6277683849, "rc"], 21.770677)
    assert subs_rc.rc.max() <= 100

    # Test alt method
    subs_rc_alt = derive_rc_alt(subs, buildings, streetcover).set_index("id")
    assert almost_equal(
        subs_rc[["area", "impervious_area", "rc"]],
        subs_rc_alt[["area", "impervious_area", "rc"]],
    )


def test_calculate_angle():
    """Test the calculate_angle function."""
    # Test with points forming a right angle
    point1 = (0, 0)
    point2 = (1, 0)
    point3 = (1, 1)
    assert go.calculate_angle(point1, point2, point3) == 90

    # Test with points forming a straight line
    point1 = (0, 0)
    point2 = (1, 0)
    point3 = (2, 0)
    assert go.calculate_angle(point1, point2, point3) == 180

    # Test with points forming an angle of 45 degrees
    point1 = (0, 0)
    point2 = (1, 0)
    point3 = (0, 1)
    assert almost_equal(go.calculate_angle(point1, point2, point3), 45)

    # Test with points forming an angle of 0 degrees
    point1 = (0, 0)
    point2 = (1, 0)
    point3 = (0, 0)
    assert go.calculate_angle(point1, point2, point3) == 0


def test_remove_intersections():
    """Test the remove_intersections function."""
    square1 = {"id": "s1", "geometry": sgeom.Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])}
    square2 = {
        "id": "s2",
        "geometry": sgeom.Polygon([(0, 0), (0.5, 0), (0.5, 0.5), (0, 0.5)]),
    }
    square3 = {
        "id": "s3",
        "geometry": sgeom.Polygon([(0.25, 0.25), (0.5, 0.25), (0.5, 0.5), (0.25, 0.5)]),
    }
    square4 = {
        "id": "s4",
        "geometry": sgeom.Polygon([(0.75, 0.75), (1, 0.75), (1, 1), (0.75, 1)]),
    }
    polys = gpd.GeoDataFrame([square1, square2, square3, square4])

    target1 = {
        "id": "s1",
        "geometry": sgeom.Polygon(
            [
                (1.0, 0.0),
                (0.5, 0.0),
                (0.5, 0.25),
                (0.5, 0.5),
                (0.25, 0.5),
                (0.0, 0.5),
                (0.0, 1.0),
                (0.75, 1.0),
                (0.75, 0.75),
                (1.0, 0.75),
                (1.0, 0.0),
            ]
        ),
    }
    target2 = {
        "id": "s2",
        "geometry": sgeom.Polygon(
            [
                (0.5, 0),
                (0, 0),
                (0, 0.5),
                (0.25, 0.5),
                (0.25, 0.25),
                (0.5, 0.25),
                (0.5, 0),
            ]
        ),
    }
    targets = gpd.GeoDataFrame([target1, target2, square3, square4])

    polys_ = go.remove_intersections(polys)

    polys_["area"] = polys_.geometry.area
    targets["area"] = targets.geometry.area
    assert polys_.set_index("id")[["area"]].equals(targets.set_index("id")[["area"]])


def test_graph_to_geojson(street_network):
    """Test the graph_to_geojson function."""
    crs = street_network.graph["crs"]
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        temp_path = Path(temp_dir)
        go.graph_to_geojson(
            street_network,
            temp_path / "graph_nodes.geojson",
            temp_path / "graph_edges.geojson",
            crs,
        )
        gdf = gpd.read_file(temp_path / "graph_nodes.geojson")
        assert gdf.crs == crs
        assert gdf.shape[0] == len(street_network.nodes)

        gdf = gpd.read_file(temp_path / "graph_edges.geojson")
        assert gdf.shape[0] == len(street_network.edges)


def test_merge_points(street_network):
    """Test the merge_points function."""
    mapping = go.merge_points(
        [(d["x"], d["y"]) for u, d in street_network.nodes(data=True)], 20
    )
    assert set(mapping.keys()) == set([2, 3, 5, 15, 16, 18, 22])
    assert set([x["maps_to"] for x in mapping.values()]) == set([2, 5, 15])
    assert mapping[15]["maps_to"] == 15
    assert mapping[18]["maps_to"] == 15
    assert almost_equal(mapping[18]["coordinate"][0], 700445.0112082)
